# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""Handles addresses of interfaces"""

import ipaddress
from ipaddress import ip_address, IPv6Address, IPv4Address

from nest.input_validator.input_validator import input_validator


class Address:
    """
    Validate IP addresses and provide some basic APIs
    to obtain information about the IP address.

    """

    @input_validator
    def __init__(self, addr_str: str):
        """
        The constructor validates the entered IP address.

        Parameters
        ----------
        addr_str : str
            The string representing the address
        """

        # Check if a "special" string is passed
        if addr_str == "DEFAULT":  # A special string in place of address
            self.ip_addr = addr_str.lower()
            return

        # Split address and prefix_len
        _addr_str, _prefix_len = self._split_address_and_prefix_length(addr_str)

        # if IPv4 address
        # expected format from user => (10.0.0.0/24 or 10.0.0.0)
        if self._is_ipv4(_addr_str):
            self.ip_addr = self._get_ipv4_address(_addr_str, _prefix_len)

        # if IPv4-Mapped IPv6 address
        # expected format from user => (::FFFF:10.0.0.0/122 or ::FFFF:10.0.0.0)
        elif self._is_ipv4_mapped_ipv6(_addr_str):
            self.ip_addr = self._get_ipv4_mapped_ipv6_address(_addr_str, _prefix_len)

        # if IPv6 address
        # expected format from user => (2a0:101::/122 or 2a0:101::)
        elif self._is_ipv6(_addr_str):
            self.ip_addr = self._get_ipv6_address(_addr_str, _prefix_len)

        else:
            raise ValueError(addr_str + " is not a valid IP address")

    @staticmethod
    def allowed_type_cast():
        """
        Indicate str can be typecasted into Address type
        """
        return [str]

    def get_addr(self, with_subnet=True):
        """Getter for ip_addr

        Parameters
        ----------
        with_subnet :
             (Default value = True)

        Returns
        -------
        str
            IP address as string
        """
        if with_subnet:
            return self.ip_addr

        return self.ip_addr.split("/")[0]

    def get_subnet(self):
        """Get the subnet of the given address"""
        if self.ip_addr == "default":
            raise Exception("default address cannot have a subnet")
        interface = ipaddress.ip_interface(self.ip_addr)
        return interface.network.compressed

    def is_subnet(self):
        """Check if the address is a subnet or not"""
        if self.ip_addr == "default":
            return False
        if self.ip_addr == self.get_subnet():
            return True
        return False

    def is_ipv6(self):
        """
        Check if the address is IPv6 or not
        Note: This takes care of 'DEFAULT' also
        """
        check_addr = self.ip_addr.split("/")[0]
        if isinstance(ip_address(check_addr), IPv6Address):
            return True
        return False

    def _split_address_and_prefix_length(self, addr_str):
        """
        Split address and prefix length

        Parameters
        ----------
        addr_str : str
            The string representing the address

        Returns
        -------
        tuple
            addr_str :str
            prefix_len : int

        """
        # raises exception in invalid case
        ipaddress.ip_interface(addr_str)

        if "/" in addr_str:
            addr_str = addr_str.split("/")
            prefix_len = int(addr_str[1])
            addr_str = addr_str[0]
        else:
            prefix_len = None

        return addr_str, prefix_len

    def _is_ipv4(self, addr_str):
        """
        Check if the address without prefix length is IPv4 or not

        Parameters
        ----------
        addr_str : str
            The string representing the address

        """
        return isinstance(ip_address(addr_str), IPv4Address)

    def _is_ipv4_mapped_ipv6(self, addr_str):
        """
        Check if the address without prefix length is IPv4-mapped IPv6 or not

        Parameters
        ----------
        addr_str : str
            The string representing the address

        """
        return ipaddress.IPv6Address(addr_str).ipv4_mapped is not None

    def _is_ipv6(self, addr_str):
        """
        Check if the address without prefix length is IPv6 or not

        Parameters
        ----------
        addr_str : str
            The string representing the address

        """
        return isinstance(ip_address(addr_str), IPv6Address)

    def _get_ipv4_address(self, addr_str, prefix_len):
        """
        Getter for IPv4 Address

        Parameters
        ----------
        addr_str : str
            The string representing the address
        prefix_len : int
            The prefix length

        Returns
        -------
        str
            IPv4 address as string

        """
        if prefix_len is None:
            prefix_len = 32
        addr_str = addr_str + "/" + str(prefix_len)
        return addr_str

    def _get_ipv4_mapped_ipv6_address(self, addr_str, prefix_len):
        """
        Getter for IPV4-mapped IPv6 Address

        Parameters
        ----------
        addr_str : str
            The string representing the address
        prefix_len : int
            The prefix length

        Returns
        -------
        str
            IPv4 address as string

        """
        addr_str = str(ipaddress.IPv6Address(addr_str).ipv4_mapped)
        # If the prefix length is empty or not valid (between 96 and 128), then it is set to 32.
        # Ipv4-mapped address has 128 bits. The IPv4 address takes the last 32 bits,
        # and therefore the first 96 (128-32) bits from the address are removed. Therefore
        # valid prefix length will be in between 96 and 128.
        if prefix_len is not None and prefix_len in range(97, 128):
            prefix_len = prefix_len - 96
        else:
            prefix_len = None
        return self._get_ipv4_address(addr_str, prefix_len)

    def _get_ipv6_address(self, addr_str, prefix_len):
        """
        Getter for IPv6 Address

        Parameters
        ----------
        addr_str : str
            The string representing the address
        prefix_len : int
            The prefix length

        Returns
        -------
        str
            IPv6 address as string

        """
        if prefix_len is None:
            prefix_len = "128"
        addr_str = addr_str + "/" + str(prefix_len)
        return addr_str

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.ip_addr!r})"


class Subnet:
    """
    For generating sequential addresses in a subnet

    """

    @input_validator
    def __init__(self, address: Address):
        if address.is_subnet():
            self._net_addr = ipaddress.ip_network(address.get_addr())
        else:
            raise ValueError("Parameter is not a subnet/network address")

        self._counter = 0

    def get_next_addr(self):
        """Get next address in sequence in the subnet"""
        self._counter += 1
        try:
            address = self._net_addr[self._counter]
        except IndexError as err:
            raise ValueError(
                "All the addresses for the network have been exhausted"
            ) from err
        return Address(address.compressed + "/" + str(self._net_addr.prefixlen))

    @property
    def counter(self):
        """Getter for counter"""
        return self._counter

    @property
    def network_address(self):
        """Getter for network address"""
        return self._net_addr

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self._net_addr!r})"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""uname commands"""

from .exec import exec_subprocess


def get_kernel_version():
    """Get Linux kernel version of the system"""
    version = exec_subprocess("uname -r", output=True)
    return version.split("-")[0]



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

""" Runs iperf command to setup UDP flows """

from time import sleep
from functools import partial
import copy
import json
import logging

from nest.experiment.interrupts import handle_keyboard_interrupt
from nest.topology_map import TopologyMap
from ..results import Iperf3Results
from .runnerbase import Runner
from ...engine.iperf3 import run_iperf_server, run_iperf_client

logger = logging.getLogger(__name__)


class Iperf3Runner(Runner):
    """
    Runs iperf client and server. Currently being used
    for UDP flows only

    Attributes
    ----------
    ns_id : str
        network namespace to run iperf from
    """

    # fmt: off

    default_iperf3_options = {
        "interval": "-i 0.2",       # Generate interim results every INTERVAL seconds
        "json": "--json",           # Get output in easily parsable JSON format
        "testlen": "-t 10",         # Length of test (NOTE: Default 10s)
        "protocol": "--udp",        # Transport protocol to use (TODO: Add support for TCP as well)
        "n_flows": "-P 1",          # Number of parallel flows (NOTE: Default 1)
        "bitrate": "-b 1M"          # Target bitrate (NOTE: Default 1 Mbit/sec)
    }

    # fmt: on

    # pylint: disable=too-many-arguments
    def __init__(
        self, ns_id, destination_ip, bandwidth, n_flows, start_time, run_time, dst_ns
    ):
        """
        Constructor to initialize the runner

        Parameters
        ----------
        ns_id : str
            network namespace to run iperf3 from
        destination : str
            the ip of server to which it has to connect
        bandwidth : int
            target bandwidth of the UDP flow in mbits
        n_flows : int
            number of parallel flows
        start_time : num
            start time of the flow
        run_time : num
            test duration
        dst_ns: str
            network namespace to run iperf3 server from
        """
        self.bandwidth = bandwidth
        self.n_flows = n_flows
        super().__init__(ns_id, start_time, run_time, destination_ip, dst_ns)

    # Should this be placed somewhere else?
    @staticmethod
    def run_server(ns_id):
        """
        Run iperf server in `ns_id`

        Parameters
        ----------
        ns_id : str
            namespace to run netserver on
        """
        return_code = run_iperf_server(ns_id)
        if return_code != 0:
            ns_name = TopologyMap.get_namespace(ns_id)["name"]
            logger.error("Error running iperf3 server at %s.", ns_name)

    def run(self):
        """
        calls engine method to run iperf client
        """
        iperf3_options = copy.copy(Iperf3Runner.default_iperf3_options)

        # Set run time
        iperf3_options["testlen"] = f"-t {self.run_time}"

        # Set number of parallel flows
        iperf3_options["n_flows"] = f"-P {self.n_flows}"

        # Set target bitrate
        iperf3_options["bitrate"] = f"-b {self.bandwidth}"

        # Convert iperf3_options dict to string
        iperf3_options_list = list(iperf3_options.values())
        iperf3_options_string = " ".join(iperf3_options_list)

        if self.start_time != 0:
            sleep(self.start_time)

        super().run(
            partial(
                run_iperf_client,
                self.ns_id,
                iperf3_options_string,
                self.destination_address.get_addr(with_subnet=False),
                self.destination_address.is_ipv6(),
            ),
            error_string_prefix="Running iperf3",
        )

    @handle_keyboard_interrupt
    def parse(self):
        """
        Parse iperf3 output from self.out
        """
        self.out.seek(0)  # rewind to start of the temp file
        raw_stats = self.out.read().decode()

        # Iperf3 already gives output in JSON format, so we parse
        # into a dictionary
        parsed_stats = json.loads(raw_stats)

        # Extract from useful connection info from parsed data
        # Useful for diffrentiating between flows
        start_timestamp = parsed_stats["start"]["timestamp"]["timesecs"]
        connection_info = {}

        for conn in parsed_stats["start"]["connected"]:
            connection_info[conn["socket"]] = {
                "local_host": conn["local_host"],
                "local_port": conn["local_port"],
                "remote_host": conn["remote_host"],
                "remote_port": conn["remote_port"],
            }

        # Convert iperf3 JSON data into our own JSON format (similar
        # to netperf), so that plotter code can be reused

        stats_dict_list = {}
        for socket in connection_info:
            local_port = connection_info[socket]["local_port"]
            stats_dict_list[local_port] = [self.get_meta_item()]

        if len(stats_dict_list) > 1:
            stats_dict_list["sum"] = [self.get_meta_item()]

        for interval in parsed_stats["intervals"]:
            for stream in interval["streams"]:
                socket = stream["socket"]
                local_port = connection_info[socket]["local_port"]
                stats_dict_list[local_port].append(
                    self._extract_from_ipref3_stream(start_timestamp, stream)
                )

            if len(stats_dict_list) > 1:
                sum_stream = interval["sum"]
                stats_dict_list["sum"].append(
                    self._extract_from_ipref3_stream(start_timestamp, sum_stream)
                )

        destination_ip = self.destination_address.get_addr(with_subnet=False)
        Iperf3Results.add_result(self.ns_id, {destination_ip: stats_dict_list})

    def _extract_from_ipref3_stream(self, start_timestamp, stream):
        """
        Convert information in iperf3 stream into required
        dictionary format
        """
        return {
            "timestamp": str(stream["start"] + start_timestamp),
            "sending_rate": str(stream["bits_per_second"] / 1e6),
            "duration": str(stream["seconds"]),
            "bytes": str(stream["bytes"]),
            "packets": str(stream["packets"]),
        }



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""Test APIs from topology sub-package"""

import unittest
import subprocess

from nest.topology import Node, connect
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap
from nest.routing.routing_helper import RoutingHelper
from nest.topology.network import Network
from nest.topology.address_helper import AddressHelper

# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-variable
# pylint: disable=consider-using-with


class TestIPv4AddressHelper(unittest.TestCase):
    def setUp(self):
        self.n0 = Node("n0")
        self.n1 = Node("n1")
        self.net1 = Network("10.0.1.0/24")
        self.net2 = Network("10.0.2.0/24")
        self.net3 = Network("2001:101::/122")
        self.net4 = Network("2002:101::/122")

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

    def test_p2p(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1, network=self.net1)

        AddressHelper.assign_addresses()

        status = self.n0.ping(n1_n0.get_address(), verbose=False)
        self.assertTrue(status)

    def test_p2p_ipv6(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1, network=self.net3)

        AddressHelper.assign_addresses()

        status = self.n0.ping(n1_n0.get_address(), verbose=False)
        self.assertTrue(status)

    def test_prp(self):
        # pylint: disable=invalid-name
        r = Node("r")
        r.enable_ip_forwarding()

        with self.net1:
            (n0_r, r_n0) = connect(self.n0, r)

        with self.net2:
            (r_n1, n1_r) = connect(r, self.n1, network=self.net2)

        AddressHelper.assign_addresses()

        self.n0.add_route("DEFAULT", n0_r)
        self.n1.add_route("DEFAULT", n1_r)

        status = self.n0.ping(n1_r.get_address(), verbose=False)

        self.assertTrue(status)

    def test_prp_ipv6(self):
        # pylint: disable=invalid-name
        r = Node("r")
        r.enable_ip_forwarding()

        with self.net3:
            (n0_r, r_n0) = connect(self.n0, r)

        with self.net4:
            (r_n1, n1_r) = connect(r, self.n1, network=self.net4)

        AddressHelper.assign_addresses()

        self.n0.add_route("DEFAULT", n0_r)
        self.n1.add_route("DEFAULT", n1_r)

        status = self.n0.ping(n1_r.get_address(), verbose=False)

        self.assertTrue(status)

    def test_prrp(self):
        # pylint: disable=invalid-name
        r1 = Node("r")
        r2 = Node("r")
        r1.enable_ip_forwarding()
        r2.enable_ip_forwarding()

        with self.net1:
            (n0_r1, r1_n0) = connect(self.n0, r1)

        (r1_r2, r2_r1) = connect(r1, r2, network=self.net2)
        (r2_n1, n1_r2) = connect(r2, self.n1)

        AddressHelper.assign_addresses()
        r2_n1.set_address("10.1.3.2/24")
        n1_r2.set_address("10.1.3.1/24")

        RoutingHelper(protocol="rip").populate_routing_tables()

        status = self.n0.ping(n1_r2.get_address(), verbose=False)

        self.assertTrue(status)

    def test_prrp_ipv6(self):
        # pylint: disable=invalid-name
        r1 = Node("r")
        r2 = Node("r")
        r1.enable_ip_forwarding()
        r2.enable_ip_forwarding()

        with self.net3:
            (n0_r1, r1_n0) = connect(self.n0, r1)

        (r1_r2, r2_r1) = connect(r1, r2, network=self.net4)
        (r2_n1, n1_r2) = connect(r2, self.n1)

        AddressHelper.assign_addresses()
        r2_n1.set_address("2003:101::10:1/122")
        n1_r2.set_address("2003:101::10:2/122")

        RoutingHelper(protocol="rip").populate_routing_tables()

        status = self.n0.ping(n1_r2.get_address(), verbose=False)

        self.assertTrue(status)

    def test_tcp_param(self):
        self.n0.configure_tcp_param("ecn", "1")
        ecn = self.n0.read_tcp_param("ecn")
        self.assertEqual(ecn, "1")

    def test_run_inside_node(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1, network=self.net1)

        AddressHelper.assign_addresses()

        # Run ping from self.n0 to self.n1
        with self.n0:
            command = f"ping -c 1 {n1_n0.address.get_addr(with_subnet=False)}"
            proc = subprocess.Popen(
                command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            (stdout, _) = proc.communicate()

        self.assertEqual(stdout[:4], b"PING", "Invalid ping output")


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
Contains classes for validation for commonly used
metrics in NeST. For eg., Bandwidth and Delay
"""

import re
from .input_validator import input_validator


class Metric:
    """
    Base type for metrics. Metrics are of the format <value><unit>.

    For eg., `Bandwidth` type ("10mbit") derives from this class. Here,
    value="10" and unit="mbit"
    """

    valid_units = []

    @input_validator
    def __init__(
        self, metric: str, value_regex=r"[0-9]+(\.[0-9]+)?", unit_regex=r"[A-Za-z]+"
    ):
        """
        Constructor for Metric. Validates if `metric` is in <value><unit>
        format.

        Parameters
        ----------
        metric: str
            Value of the metric
        """
        if not re.fullmatch(value_regex + unit_regex, metric):
            raise ValueError(
                f"{metric} is not a valid metric. Metrics must be of the format "
                f"<value><unit>"
            )

        # Extract value and unit
        self._value = re.search(value_regex, metric).group()
        self._unit = re.search(unit_regex, metric).group()

        self._string_value = metric

    @property
    def value(self):
        """
        Get the value of the metric (without unit)
        """
        return self._value

    @property
    def unit(self):
        """
        Get the unit of the metric
        """
        return self._unit

    @property
    def string_value(self):
        """
        Get string value of the mrtic
        """
        return self._string_value

    @staticmethod
    def allowed_type_cast():
        """
        Indicate str can be typecasted into Metric type
        """
        return [str]


class Bandwidth(Metric):
    """
    Validates that the bandwidth is in expected format.
    """

    # fmt: off

    valid_units = [
        "bit", "kbit", "kibit", "mbit", "mibit", "gbit", "gibit", "tbit", "tibit",
        "bps", "kbps", "kibps", "mbps", "mibps", "gbps", "gibps", "tbps", "tibps",
    ]

    # fmt: on

    def __init__(self, bandwidth: str):
        """
        Parameters
        ----------
        bandwidth: str
            The value of bandwidth
        """
        super().__init__(bandwidth)

        if self.unit not in Bandwidth.valid_units:
            raise ValueError(f"{self.unit} is not a valid unit for bandwidth.")


class Delay(Metric):
    """
    Validates that the delay is in expected format.
    """

    valid_units = ["s", "sec", "secs", "ms", "msec", "msecs", "us", "usec", "usecs"]

    def __init__(self, delay: str):
        """
        Parameters
        ----------
        delay: str
            The value of delay
        """
        super().__init__(delay)

        if self.unit not in Delay.valid_units:
            raise ValueError(f"{self.unit} is not a valid unit for delay.")


class Percentage(Metric):
    """
    Validates that the percentage is in expected format.
    """

    def __init__(self, percentage: str):
        """
        Parameters
        ----------
        percentage: str
            The value of percentage
        """
        super().__init__(percentage, unit_regex=r"%")

        if not 0 <= float(self.value) <= 100:
            raise ValueError(f"{self.string_value} is not a valid percentage.")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Handles collection of results (raw data)"""

import json
from multiprocessing import Manager
from ..topology_map import TopologyMap
from .pack import Pack


class Results:
    """This class aggregates the stats from the entire experiment environment"""

    @staticmethod
    def add_result(results_q, ns_id, result):
        """
        Adds the stats parse from a process to the shared `results_q`

        Parameters
        ----------
        results_q : multiprocessing.Manager.Queue
            Shared stats
        ns_id : string
            namespace id (internal name)
        result : dict
            parsed stats
        """
        # Convert nest's internal name to user given name
        ns_name = TopologyMap.get_namespace(ns_id)["name"]

        item = results_q.get()
        if ns_name not in item:
            item[ns_name] = [result]
        else:
            temp = item[ns_name]
            temp.append(result)
            item[ns_name] = temp
        results_q.put(item)

    @staticmethod
    def remove_all_results(results_q):
        """
        Remove all results obtained from the experiment

        Parameters
        ----------
        results_q : multiprocessing.Manager.Queue
            Shared stats
        """
        results_q.get()
        results_q.put({})

    @staticmethod
    def get_results(results_q):
        """
        Get results obtained in the experiment so far

        Parameters
        ----------
        results_q : multiprocessing.Manager.Queue
            Shared stats
        """
        results = results_q.get()
        results_q.put(results)
        return results

    @staticmethod
    def output_to_file(results_q, toolname):
        """
        Outputs the aggregated results into a file.
        If results are empty, then it is not output to file.

        Parameters
        ----------
        results_q : multiprocessing.Manager.Queue
            Shared stats
        toolname : str
            Like ss, tc, netperf
        """
        results = Results.get_results(results_q)
        if results:
            json_stats = json.dumps(results, indent=4)
            Pack.dump_file("{}.json".format(toolname), json_stats)


# Shared variables to aggregate results
ss_results_q = Manager().Queue()
ss_results_q.put({})


class SsResults:
    """This class aggregates the ss stats from the entire experiment environment"""

    @staticmethod
    def add_result(ns_id, result):
        """Adds the ss stats parse from a process to the shared `ss_results`

        Parameters
        ----------
        ns_id : string
            namespace id (internal name)
        result : dict
            parsed ss stats
        """
        Results.add_result(ss_results_q, ns_id, result)

    @staticmethod
    def remove_all_results():
        """Remove all results obtained from the experiment"""
        Results.remove_all_results(ss_results_q)

    @staticmethod
    def get_results():
        """Get results obtained in the experiment so far"""
        return Results.get_results(ss_results_q)

    @staticmethod
    def output_to_file():
        """Outputs the aggregated ss stats to file"""
        Results.output_to_file(ss_results_q, "ss")


# Shared variables to aggregate results
netperf_results_q = Manager().Queue()
netperf_results_q.put({})


class NetperfResults:
    """This class aggregates the netperf stats from the entire experiment environment"""

    @staticmethod
    def add_result(ns_id, result):
        """Adds the netperf stats parse from a process to the shared `netperf_results`

        Parameters
        ----------
        ns_id : string
            namespace id (internal name)
        result : dict
            parsed netperf stats
        """
        Results.add_result(netperf_results_q, ns_id, result)

    @staticmethod
    def remove_all_results():
        """Remove all results obtained from the experiment"""
        Results.remove_all_results(netperf_results_q)

    @staticmethod
    def get_results():
        """Get results obtained in the experiment so far"""
        return Results.get_results(netperf_results_q)

    @staticmethod
    def output_to_file():
        """Outputs the aggregated netperf stats to file"""
        Results.output_to_file(netperf_results_q, "netperf")


# Shared variables to aggregate results
iperf3_results_q = Manager().Queue()
iperf3_results_q.put({})


class Iperf3Results:
    """This class aggregates the iperf3 stats from the entire experiment environment"""

    @staticmethod
    def add_result(ns_id, result):
        """Adds the iperf3 stats parsed from a process to the shared `iperf3_results`

        Parameters
        ----------
        ns_id : string
            namespace id (internal name)
        result : dict
            parsed netperf stats
        """
        Results.add_result(iperf3_results_q, ns_id, result)

    @staticmethod
    def remove_all_results():
        """Remove all results obtained from the experiment"""
        Results.remove_all_results(iperf3_results_q)

    @staticmethod
    def get_results():
        """Get results obtained in the experiment so far"""
        return Results.get_results(iperf3_results_q)

    @staticmethod
    def output_to_file():
        """Outputs the aggregated netperf stats to file"""
        Results.output_to_file(iperf3_results_q, "iperf3")


# Shared variables to aggregate results
tc_results_q = Manager().Queue()
tc_results_q.put({})


class TcResults:
    """This class aggregates the tc stats from the entire experiment environment"""

    @staticmethod
    def add_result(ns_id, result):
        """Adds the tc stats parse from a process to the shared `tc_results`

        Parameters
        ----------
        ns_id : string
            namespace id (internal name)
        result : dict
            parsed tc stats
        """
        Results.add_result(tc_results_q, ns_id, result)

    @staticmethod
    def remove_all_results():
        """Remove all results obtained from the experiment"""
        Results.remove_all_results(tc_results_q)

    @staticmethod
    def get_results():
        """Get results obtained in the experiment so far"""
        return Results.get_results(tc_results_q)

    @staticmethod
    def output_to_file():
        """Outputs the aggregated tc stats to file"""
        Results.output_to_file(tc_results_q, "tc")


ping_results_q = Manager().Queue()
ping_results_q.put({})


class PingResults:
    """This class aggregates the ping stats from the entire experiment environment"""

    @staticmethod
    def add_result(ns_id, result):
        """Adds the ping stats parse from a process to the shared `ping_results`

        Parameters
        ----------
        ns_id : string
            namespace id (internal name)
        result : dict
            parsed ping stats
        """
        Results.add_result(ping_results_q, ns_id, result)

    @staticmethod
    def remove_all_results():
        """Remove all results obtained from the experiment"""
        Results.remove_all_results(ping_results_q)

    @staticmethod
    def get_results():
        """Get results obtained in the experiment so far"""
        return Results.get_results(ping_results_q)

    @staticmethod
    def output_to_file():
        """Outputs the aggregated ping stats to file"""
        Results.output_to_file(ping_results_q, "ping")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Common plotting logic among tools"""

import numpy as np
import matplotlib.pyplot as plt

# pylint: disable=invalid-name

# pylint: disable=too-many-arguments
def simple_plot(title, x_list, y_list, x_label, y_label, legend_string=None):
    """
    Plot values

    Parameters
    ----------
    title : str
        Title for the plot
    x_list : List
        List of x values
    y_list : List
        List of y values
    x_label : str
        Label for x values
    y_label : str
        Label for y values
    legend_string : None/str
        If a string, then have a legend in the plot

    Returns
    -------
    matplotlib.plt.fig
        fig of plot
    """
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot(x_list, y_list)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_title(title)

    if legend_string is not None:
        ax.legend([legend_string])

    return fig


def mix_plot(title, data, x_label, y_label, with_sum=False):
    """Plot multiple sets of values and their total sum

    Parameters
    ----------
    title : str
        Title for the plot
    data : List
        list of values
    x_label : str
        Label for x values
    y_label : str
        Label for y values
    with_sum : boolean
        If should plot the sum of all y values (Default value = False)

    Returns
    -------
    maplotlib.plt.fig
        fig of plot
    """
    fig, ax = plt.subplots()
    # TODO: I couldn't think of better variable name :(
    for chunk in data:
        (x_list, y_list) = chunk["values"]
        label = chunk["label"]
        ax.plot(x_list, y_list, label=label)

    if with_sum:

        x = []
        for chunk in data:
            (x_list, _) = chunk["values"]
            x.append(x_list)

        # Get sorted list of all x values
        x = np.unique(np.concatenate(x))

        total = np.array([0.0] * len(x))
        for chunk in data:
            (x_list, y_list) = chunk["values"]
            # Interpolate y values on the combined x values
            y = np.interp(x, x_list, y_list, left=0, right=0)
            total += y

        ax.plot(x, total, label="Aggregate", alpha=0.5)

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_title(title)
    ax.legend()

    return fig



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Plot netperf results"""

import logging
import matplotlib.pyplot as plt
from nest.experiment.interrupts import handle_keyboard_interrupt
from ..pack import Pack
from .common import simple_plot, mix_plot

logger = logging.getLogger(__name__)


def _plot_netperf_flow(flow, node, dest):
    """
    Plot netperf stats of the flow

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    flow : List
        List with timestamps and stats
    node : str
        Node from which netperf results were obtained from
    dest :
        Destination ip:port address of the flow

    Returns
    -------
    tuple/None
        Timestamped sending_rate values
    """
    # "meta" item will always be present, hence `<= 1`
    if len(flow) <= 1:
        logger.warning(
            "Flow from %s to destination %s " "doesn't have any parsed netperf result.",
            node,
            dest,
        )
        return None

    # First item is the "meta" item with user given information
    user_given_start_time = float(flow[0]["start_time"])
    destination_node = flow[0]["destination_node"]

    # "Bias" actual start_time in experiment with user given start time
    start_time = float(flow[1]["timestamp"]) - user_given_start_time

    timestamp = []
    sending_rate = []

    for data in flow[1:]:
        sending_rate.append(float(data["sending_rate"]))
        relative_time = float(data["timestamp"]) - start_time
        timestamp.append(relative_time)

    legend_string = f"{node} to {destination_node} ({dest})"

    # TODO: Check if sending_rate is always in Mbps
    fig = simple_plot(
        "",
        timestamp,
        sending_rate,
        "Time (Seconds)",
        "Sending Rate (Mbps)",
        legend_string=legend_string,
    )

    filename = f"sending_rate_{node}_to_{destination_node}({dest}).png"
    Pack.dump_plot("netperf", filename, fig)
    plt.close(fig)

    return {"label": legend_string, "values": (timestamp, sending_rate)}


@handle_keyboard_interrupt
def plot_netperf(parsed_data):
    """
    Plot statistics obtained from netperf

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    parsed_data : Dict
        JSON data parsed from ss
    """
    for node in parsed_data:
        node_data = parsed_data[node]

        all_flow_data = []
        for connection in node_data:
            for dest in connection:
                flow = connection[dest]
                plotted_data = _plot_netperf_flow(flow, node, dest)
                if plotted_data is not None:
                    all_flow_data.append(plotted_data)

        if len(all_flow_data) > 1:
            fig = mix_plot(
                "",
                all_flow_data,
                "Time (Seconds)",
                "Sending Rate (Mbps)",
                with_sum=True,
            )
            filename = f"sending_rate_{node}.png"
            Pack.dump_plot("netperf", filename, fig)
            plt.close(fig)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""
NeST internally uses auto-generated unique ids to name topologies.
This module holds the mapping between the user given names and the names
generated internally by NeST.
"""

import json
import logging

logger = logging.getLogger(__name__)


class TopologyMap:
    """
    Store mapping between user given names and NeST ids
    """

    # topology_map contains the information about topology created
    topology_map = {"namespaces": [], "hosts": [], "routers": []}
    # NOTE:
    # To figure out the contents of each dict keyword,
    # checkout the add_keyword function.
    # For example, to get the contents of qdisc parameter,
    # checkout add_qdisc function

    # Pointer to topology_map['namespaces']
    # Used for efficiency
    # orphan interfaces are those interfaces which do not belong to any network
    # list of network is used to store all the existing network object
    namespaces_pointer = {}
    list_of_network = []
    orphan_interfaces = 0

    @staticmethod
    def add_namespace(ns_id, ns_name):
        """
        Add namespace to topology_map

        Parameters
        ----------
        ns_id : str
            namepspace id
        ns_name : str
            namespace name
        """
        if ns_id in TopologyMap.namespaces_pointer:
            raise ValueError(f"Namespace with id {ns_id} already exists in TopologyMap")

        namespaces = TopologyMap.get_namespaces()

        namespaces.append({"id": ns_id, "name": ns_name, "interfaces": []})

        TopologyMap.namespaces_pointer[ns_id] = {
            "pos": len(namespaces) - 1,
            "interfaces_pointer": {},
        }

    @staticmethod
    def add_host(host):
        """
        Add host to topology_map. This is required apart
        from `add_namespaces` for routing

        Parameters
        ----------
        ns_id : str
            namepspace id of the host
        ns_name : str
            namespace name of the host
        """
        hosts = TopologyMap.get_hosts()

        hosts.append(host)

    @staticmethod
    def add_router(router):
        """
        Add router to topology_map. This is required apart
        from `add_namespaces` for routing

        Parameters
        ----------
        ns_id : str
            namepspace id of the router
        ns_name : str
            namespace name of the router
        """
        routers = TopologyMap.get_routers()

        routers.append(router)
        TopologyMap.get_hosts().remove(router)  # remove the router from hosts list

    @staticmethod
    def add_interface(ns_id, int_id, int_name):
        """
        Add interface to topology_map

        Parameters
        ----------
        ns_id : str
            namepspace id
        int_id : str
            interface id
        int_name :
            interface name
        """
        if ns_id not in TopologyMap.namespaces_pointer:
            raise ValueError(f"Namespace with id {ns_id} doesn't exist in TopologyMap")

        if int_id in TopologyMap.namespaces_pointer[ns_id]["interfaces_pointer"]:
            raise ValueError(
                f"Interface with id {int_id} already present in namespace {ns_id}"
            )

        # TODO: classes not added yet to list
        interfaces = TopologyMap.get_interfaces(ns_id)

        interfaces.append({"id": int_id, "name": int_name, "qdiscs": []})

        TopologyMap.orphan_interfaces = TopologyMap.orphan_interfaces + 1

        TopologyMap.namespaces_pointer[ns_id]["interfaces_pointer"][int_id] = {
            "pos": len(interfaces) - 1
        }

    @staticmethod
    def add_qdisc(ns_id, int_id, kind, handle, parent=""):
        """
        Add qdisc to topology_map

        Parameters
        ----------
        ns_id : str
            namepspace id
        int_id : str
            interface id
        kind : str
            qdisc kind
        handle : str
            qdisc handle
        parent : str
            qdisc parent (Default value = '')
        """
        if ns_id not in TopologyMap.namespaces_pointer:
            raise ValueError(f"Namespace with id {ns_id} doesn't exist in TopologyMap")

        if int_id not in TopologyMap.namespaces_pointer[ns_id]["interfaces_pointer"]:
            raise ValueError(
                f"Interface with id {int_id} doesn't exist in namespace {ns_id}"
            )

        # TODO: Check if qdisc is already present?

        qdiscs = TopologyMap.get_qdiscs(ns_id, int_id)
        qdiscs.append({"kind": kind, "handle": handle, "parent": parent})

    @staticmethod
    def change_qdisc(ns_id, int_id, kind, handle):
        """
        Add qdisc to topology_map

        Parameters
        ----------
        ns_id : str
            namepspace id
        int_id : str
            interface id
        kind : str
            qdisc kind
        handle : str
            qdisc handle
        parent : str
            qdisc parent (Default value = '')
        """
        if ns_id not in TopologyMap.namespaces_pointer:
            raise ValueError(f"Namespace with id {ns_id} doesn't exist in TopologyMap")

        if int_id not in TopologyMap.namespaces_pointer[ns_id]["interfaces_pointer"]:
            raise ValueError(
                f"Interface with id {int_id} doesn't exist in namespace {ns_id}"
            )

        qdiscs = TopologyMap.get_qdiscs(ns_id, int_id)
        for qdisc in qdiscs:
            if qdisc["handle"] == handle:
                qdisc["kind"] = kind

    @staticmethod
    def delete_qdisc(ns_id, int_id, handle):
        """
        Delete qdisc from topology_map

        Parameters
        ----------
        ns_id : str
            namepspace id
        int_id : str
            interface id
        handle : str
            qdisc handle
        """
        qdiscs = TopologyMap.get_qdiscs(ns_id, int_id)
        counter = 0
        for qdisc in qdiscs:
            if qdisc["handle"] == handle:
                qdiscs.pop(counter)
                break
            counter += 1

    @staticmethod
    def get_topology_map():
        """Get the entire topology map"""
        return TopologyMap.topology_map

    @staticmethod
    def get_namespaces():
        """
        Get all namespaces added

        Returns
        -------
        List[Dict]
        """
        return TopologyMap.topology_map["namespaces"]

    @staticmethod
    def get_routers():
        """
        Get all routers added

        Returns
        -------
        List[Dict]
        """
        return TopologyMap.topology_map["routers"]

    @staticmethod
    def get_hosts():
        """
        Get all hosts added

        Returns
        -------
        List[Dict]
        """
        return TopologyMap.topology_map["hosts"]

    @staticmethod
    def get_namespace(ns_id, with_interfaces_pointer=False):
        """
        Get namespace given its id

        Parameters
        ----------
        ns_id : str
            namespace id
        with_interfaces_pointer : bool
            If should return interfaces_pointer for the namespace (Default value = False)

        Returns
        -------
        Dict
            Required namespace with id, name and interfaces in it
        """
        namespaces = TopologyMap.get_namespaces()
        namespace_pointer = TopologyMap.namespaces_pointer[ns_id]
        namespace = namespaces[namespace_pointer["pos"]]

        if with_interfaces_pointer:
            return (namespace_pointer["interfaces_pointer"], namespace)

        return namespace

    @staticmethod
    def get_interfaces(ns_id):
        """
        Get all interfaces in the namespace

        Parameters
        ----------
        ns_id : str
            namespace id

        Returns
        -------
            List[Dict]
        """
        namespace = TopologyMap.get_namespace(ns_id)
        interfaces = namespace["interfaces"]

        return interfaces

    @staticmethod
    def get_interface(ns_id, int_id):
        """
        Get interface in namespace `ns_id` with interface
        `int_id`

        Parameters
        ----------
        ns_id : str
            namespace id
        int_id : str
            interface id

        Returns
        -------
        Dict
            Interface details
        """
        (interfaces_pointer, _) = TopologyMap.get_namespace(
            ns_id, with_interfaces_pointer=True
        )
        interfaces = TopologyMap.get_interfaces(ns_id)
        interface_pointer = interfaces_pointer[int_id]
        interface = interfaces[interface_pointer["pos"]]

        return interface

    @staticmethod
    def get_qdiscs(ns_id, int_id):
        """
        Get qdiscs in namespace `ns_id` with interface
        `int_id`

        Parameters
        ----------
        ns_id : str
            namespace id
        int_id : str
            interface id

        Returns
        -------
            List[Dict]
        """
        interface = TopologyMap.get_interface(ns_id, int_id)
        qdiscs = interface["qdiscs"]

        return qdiscs

    @staticmethod
    def delete_all_mapping():
        """
        Delete all mappings stored in `TopologyMap`.
        """
        TopologyMap.topology_map = {"namespaces": [], "hosts": [], "routers": []}
        TopologyMap.namespaces_pointer = {}
        TopologyMap.list_of_network = []
        TopologyMap.orphan_interfaces = 0

    @staticmethod
    def dump():
        """
        Dump generated topology_map. (for debugging purposes)
        """
        logger.debug("Config")
        logger.debug("------")
        logger.debug(json.dumps(TopologyMap.topology_map, indent=4))

        # print()
        # print('Pointers')
        # print('--------')
        # print(json.dumps(TopologyMap.namespaces_pointer, indent = 4))

    @staticmethod
    def add_network(network):
        """
        Add network object reference to topology_map

        Parameters
        ----------
        network : Network
            Object reference of Network class
        """
        TopologyMap.list_of_network.append(network)

    @staticmethod
    def decrement_orphan_interfaces():
        """
        Reduce the orphan interface in the topology

        """
        TopologyMap.orphan_interfaces = TopologyMap.orphan_interfaces - 1



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Topology module
===============

This module is responsible for setting up an emulated topology.

In topology creation,

* Nodes are created
* Connections are made between nodes
* Addresses are assigned to interfaces
* Bandwidth and latency are set
* Qdiscs are installed
"""

import uuid

from . import id_generator
from .address import Address, Subnet
from .node import Node
from .router import Router
from .interface import Interface, connect
from .switch import Switch


# Generate unique topology id for the *to be created* topology
TOPOLOGY_ID = uuid.uuid4().hex[:10]  # TODO: First 10 seems hacky
id_generator.IdGen(TOPOLOGY_ID)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""IPv6 States Check"""

from .exec import exec_subprocess


def check_ipv6_states(namespaces):
    """
    Checks IPv6 states

    Parameters
    ----------
    ns_name : str
        namespace name

    Returns
    -------
    dict
        Returns dictionary of states for IPv6 addresses
    """

    states = {
        "tentative": False,
        "dadfailed": [
            False,
        ],
    }

    # Checks in all the namespaces if any interface is in tentative state or not
    for i in range(len(namespaces)):
        status = exec_subprocess(
            f"ip netns exec {namespaces[i]['id']} ip addr show | grep -c tentative",
            shell=True,
            output=True,
        )

        if int(status) != 0:
            states["tentative"] = True
            break

    # Checks in all the namespaces if any interface is in dadfailed state or not
    for i in range(len(namespaces)):
        status = exec_subprocess(
            f"ip netns exec {namespaces[i]['id']} ip addr show | grep -c dadfailed",
            shell=True,
            output=True,
        )

        if int(status) != 0:
            states["dadfailed"][0] = True
            states["dadfailed"].append(namespaces[i]["name"])
            break

    return states



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Handle plotting results obtained"""
import matplotlib
from matplotlib import style

# Since plots are not shown in window, use the non-interactive Agg backend.
# Also fixes https://gitlab.com/nitk-nest/nest/-/issues/73
# Many plotting processes were connecting to X-server simultaneously.
# This created errors as X-server couldn't handle requests
matplotlib.use("agg")

# Set plot style
style.use("seaborn-paper")
style.use("ggplot")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to device creation in topology"""

import logging
import nest.config as config
from nest.topology_map import TopologyMap
from nest.topology.id_generator import IdGen
from nest.topology.traffic_control import TrafficControlHandler
from nest import engine

logger = logging.getLogger(__name__)

# Max length of device when Topology Map is disabled
# i.e. 'assign_random_names' is set to False in config
MAX_CUSTOM_NAME_LEN = 15


class Device:
    """
    A common subclass which all devices inherit

    Attributes
    ----------

    name : str
        User given name for the device
    id : str
        This value is used by `engine` to create emulated device
        entity
    node_id : str
        This is id of the node that the device belongs to
    traffic_ctl_hanlder : TrafficControlHanlder
        Takes care of traffic control for the device
    """

    def __init__(self, name, node_id):
        """
        Constructore for a device

        Parameters
        ----------
        name : str
            User given name for the device
        node_id : str
            This is the id of the node that the device belongs to
        """

        if config.get_value("assign_random_names") is False:
            if len(name) > MAX_CUSTOM_NAME_LEN:
                raise ValueError(
                    f"Device name {name} is too long. Device names "
                    f"should not exceed 15 characters"
                )

        self._name = name
        self._id = IdGen.get_id(name)
        self._traffic_control_handler = TrafficControlHandler(node_id, self._id)
        if node_id is not None:
            TopologyMap.add_interface(self.node_id, self._id, self._name)

    @property
    def name(self):
        """Getter for name"""
        return self._name

    @property
    def id(self):
        """Getter for id"""
        return self._id

    @property
    def qdisc_list(self):
        """
        Getther for the list of all qdiscs
        """
        return self._traffic_control_handler.qdisc_list

    @property
    def node_id(self):
        """
        Getter for the `Node` associated
        with the device
        """
        return self._traffic_control_handler.node_id

    @node_id.setter
    def node_id(self, node_id):
        """
        Setter for the `Node` associated
        with the device

        Parameters
        ----------
        node : str
            The id of the node where the device is to be installed
        """
        self._traffic_control_handler.node_id = node_id
        if node_id is not None:
            TopologyMap.add_interface(self.node_id, self._id, self._name)

    def set_mode(self, mode):
        """
        Changes the mode of the device

        Parameters
        ----------
        mode : string
            device mode to be set
        """
        if mode in ("UP", "DOWN"):
            if self.node_id is not None:
                engine.set_interface_mode(self.node_id, self.id, mode.lower())
            else:
                # TODO: Create our own error class
                raise NotImplementedError(
                    "You should assign the device to node or router before setting it's mode"
                )
        else:
            raise ValueError(
                f'{mode} is not a valid mode (it has to be either "UP" or "DOWN")'
            )

    def add_qdisc(self, qdisc, parent="root", handle="", **kwargs):
        """
        Add a qdisc (Queueing Discipline) to this device

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        dev : Interface class
            The device to which the qdisc is to be added
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        handle : string
            id of the filter (Default value = '')
        """

        self._traffic_control_handler.add_qdisc(qdisc, parent, handle, **kwargs)

    def change_qdisc(self, handle, qdisc="", **kwargs):
        """
        Change a qdisc that is already present in the device

        Parameters
        ----------
        handle : string
            Hande of the qdisc to be changed
        qdisc : string
            The new qdisc to be added to the device
        """

        self._traffic_control_handler.change_qdisc(handle, qdisc, **kwargs)

    def delete_qdisc(self, handle):
        """
        Delete qdisc (Queueing Discipline) from this device

        Parameters
        ----------
        handle : string
            Handle of the qdisc to be deleted
        """

        self._traffic_control_handler.delete_qdisc(handle)

    def add_class(self, qdisc, parent="root", classid="", **kwargs):
        """
        Create an object that represents a class

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        classid : string
            id of the class (Default value = '')
        """

        self._traffic_control_handler.add_class(qdisc, parent, classid, **kwargs)

    def change_class(self, qdisc, parent, classid, **kwargs):
        """
        Change a class that is already present in the device

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        classid : string
            id of the class (Default value = '')
        """

        self._traffic_control_handler.change_class(qdisc, parent, classid, **kwargs)

    def delete_class(self, classid, parent):
        """
        Delete Class from this device

        Parameters
        ----------
        handle : string
            Handle of the class to be deleted
        """
        self._traffic_control_handler.delete_class(classid, parent)

    # pylint: disable=too-many-arguments
    def add_filter(
        self,
        priority,
        filtertype,
        flowid,
        protocol="ip",
        parent="root",
        handle="",
        **kwargs,
    ):
        """
        Design a Filter to assign to a Class or Qdisc

        Parameters
        ----------
        protocol : string
            protocol used (Default value = 'ip')
        priority : int
            priority of the filter
        filtertype : string
            one of the available filters
        flowid : Class
            classid of the class where the traffic is enqueued
            if the traffic passes the filter
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        handle : string
            id of the filter (Default value = '')
        filter : dictionary
            filter parameters
        """
        # TODO: Verify type of parameters
        # TODO: Reduce arguments to the engine functions by finding parent and handle automatically

        self._traffic_control_handler.add_filter(
            priority, filtertype, flowid, protocol, parent, handle, **kwargs
        )

    def delete_filter(self, handle, parent):
        """
        Delete filter from a Class or a Qdisc

        Parameters
        ----------
        handle : string
            Handle of the class to be deleted
        """

        self._traffic_control_handler.delete_filter(handle, parent)

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Experiment Module
=================

Handles running network experiments on topologies built using
`Topology` module
"""

from .experiment import Experiment, Flow



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""User API to setup and run experiments on a given topology"""

import copy
import logging
from nest.input_validator.metric import Bandwidth
from nest.network_utilities import ipv6_dad_check
from nest.input_validator import input_validator
from nest.topology import Node, Address, Interface
from .run_exp import run_experiment
from .pack import Pack

logger = logging.getLogger(__name__)

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods
class Flow:
    """Defines a flow in the topology"""

    # pylint: disable=too-many-arguments
    @input_validator
    def __init__(
        self,
        source_node: Node,
        destination_node: Node,
        destination_address: Address,
        start_time: int,
        stop_time: int,
        number_of_streams: int,
    ):
        """
        'Flow' object in the topology

        Parameters
        ----------
        source_node : Node
            Source node of flow
        destination_node : Node
            Destination node of flow
        destination_address : Address/str
            Destination address of flow
        start_time : int
            Time to start flow (in seconds)
        stop_time : int
            Time to stop flow (in seconds)
        number_of_streams : int
            Number of streams in the flow
        """
        self.source_node = source_node
        self.destination_node = destination_node
        self.destination_address = destination_address
        self.start_time = start_time
        self.stop_time = stop_time
        self.number_of_streams = number_of_streams

        self._options = {"protocol": "TCP", "cong_algo": "cubic"}

    @property
    def destination_address(self):
        """Getter for destination address"""
        return self._destination_address

    @destination_address.setter
    def destination_address(self, destination_address):
        """Setter for destination address"""
        if isinstance(destination_address, str):
            destination_address = Address(destination_address)
        self._destination_address = destination_address

    def _get_props(self):
        """
        Get flow properties.

        NOTE: To be used internally
        """

        return [
            self.source_node.id,
            self.destination_node.id,
            self.destination_address.get_addr(with_subnet=False),
            self.start_time,
            self.stop_time,
            self.number_of_streams,
            self._options,
        ]

    def __repr__(self):
        classname = self.__class__.__name__
        return (
            f"{classname}({self.source_node!r}, {self.destination_node!r},"
            f" {self.destination_address!r}), {self.start_time!r}, {self.stop_time!r}"
            f" {self.number_of_streams!r})"
        )


class Experiment:
    """Handles experiment to be run on topology"""

    @input_validator
    def __init__(self, name: str):
        """
        Create experiment

        Parameters
        ----------
        name : str
            Name of experiment
        """
        self.name = name
        self.flows = []
        self.node_stats = []
        self.qdisc_stats = []

    def add_flow(self, flow):
        """
        Add flow to experiment
        By default, the flow is assumed to be
        TCP with cubic congestion algorithm

        Parameters
        ----------
        flow : Flow
            Add flow to experiment
        """
        self.flows.append(copy.deepcopy(flow))

    @input_validator
    def add_tcp_flow(self, flow: Flow, congestion_algorithm="cubic"):
        """
        Add TCP flow to experiment. If no congestion control algorithm
        is specified, then by default cubic is used.

        Note: The congestion control algorithm specified in this API
        overrides the congestion control algorithm specified in
        `topology.Node.configure_tcp_param()` API.

        Parameters
        ----------
        flow : Flow
            Flow to be added to experiment
        congestion_algorithm : str
            TCP congestion algorithm (Default value = 'cubic')
        """

        # TODO: Verify congestion algorithm

        options = {"protocol": "TCP", "cong_algo": congestion_algorithm}

        flow._options = options  # pylint: disable=protected-access
        self.add_flow(flow)

    @input_validator
    def add_udp_flow(
        self, flow: Flow, target_bandwidth: Bandwidth = Bandwidth("1mbit")
    ):
        """
        Add UDP flow to experiment

        Parameters
        ----------
        flow : Flow
            Flow to be added to experiment
        target_bandwidth :
            UDP bandwidth (in Mbits) (Default value = '1mbit')
            This bandwidth limit is for each UDP stream in the flow
        """
        options = {"protocol": "UDP", "target_bw": target_bandwidth.string_value}

        flow._options = options  # pylint: disable=protected-access
        self.add_flow(flow)

    @input_validator
    def require_qdisc_stats(self, interface: Interface, stats=""):
        """
        Stats to be obtained from qdisc in interface

        Parameters
        ----------
        interface : Interface
            Interface containing the qdisc
        stats : list(str)
            Stats required (Default value = '') [NOT SUPPORTED]
        """
        # TODO: Leads to rewrite if the function is called
        # twice with same 'interface'

        # for stat in stats:
        #     if stat not in Experiment.qdisc_stats:
        #         raise ValueError('{} is not a valid Queue property.'.format(stat))

        if interface.get_qdisc() is None:
            raise ValueError("Given interface hasn't been assigned any qdisc.")

        self.qdisc_stats.append(
            {
                "ns_id": interface.node_id,
                "int_id": interface.ifb_id,
                "qdisc": interface.get_qdisc().qdisc,
                "stats": stats,
            }
        )

    @ipv6_dad_check
    def run(self):
        """Run the experiment"""
        print()
        logger.info("Running experiment %s ", self.name)
        Pack.init(self.name)
        run_experiment(self)

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Plot tc results"""

import logging
import matplotlib.pyplot as plt

from .common import simple_plot
from ..pack import Pack

logger = logging.getLogger(__name__)


def _extract_from_tc_stats(stats, node, interface):
    """
    Extract information from tc stats and convert it to
    conveniently plottable data format

    Parameters
    ----------
    stats : List
        List of timestamps and stats
    node : str
        Node from which tc results were obtained from
    interface : str
        Interface from which results were obtained from
    """
    if len(stats) == 0:
        logger.warning(
            "Qdisc at %s of %s doesn't have any " "parsed tc result.", interface, node
        )
        return None

    qdisc = stats[0]["kind"]
    start_time = float(stats[0]["timestamp"])

    timestamp = []
    stats_params = {}

    for param in stats[0]:
        if param not in ("timestamp", "kind"):
            stats_params[param] = []

    for data in stats:
        for param in stats_params:
            stats_params[param].append(data[param])
        relative_time = float(data["timestamp"]) - start_time
        timestamp.append(relative_time)

    return (qdisc, timestamp, stats_params)


def _plot_tc_stats(stats, node, interface):
    """
    Plot tc stats of the flow

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    stats : List
        List with timestamps and stats
    node : str
        Node from which tc results were obtained from
    interface : str
        Interface from which results were obtained from
    """
    values = _extract_from_tc_stats(stats, node, interface)
    if values is None:
        return
    (qdisc, timestamp, stats_params) = values

    for param in stats_params:
        fig = simple_plot(
            "Traffic Control (tc) Statistics",
            timestamp,
            stats_params[param],
            "Time (Seconds)",
            param,
            legend_string=f"Interface {interface} in {node}",
        )
        filename = f"{node}_{interface}_{qdisc}_{param}.png"
        Pack.dump_plot("tc", filename, fig)
        plt.close(fig)


def plot_tc(parsed_data):
    """
    Plot statistics obtained from tc

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    parsed_data : Dict
        JSON data parsed from ss
    """
    for node in parsed_data:
        node_data = parsed_data[node]
        for interfaces in node_data:
            for interface in interfaces:
                qdisc = interfaces[interface]
                for handle in qdisc:
                    stats = qdisc[handle]
                    _plot_tc_stats(stats, node, interface)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Handle traffic control entities"""

from nest.topology_map import TopologyMap
from .. import engine

# TODO: Improve this module such that the below pylint disables are no
# longer required

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes


class TrafficControlHandler:
    """Handles all traffic control related methods for a device

    Attributes
    ----------
    node_id : str
        The id of the namespace to which it belongs
    dev_id : str
        The id of the device to to which it belongs
    qdisc_list : list
        The list of all qdisc on a specific device
    class_list : list
        The list of all class on a specific device
    filter_list : list
        The list of all filters on a specific device
    """

    def __init__(self, node_id, dev_id):
        """
        Constructor for a traffic controller

        Parameters
        ----------
        node_id : str
            The id of the namespace to which the device belongs to
        dev_id : str
            The id of the device to which the traffic controller belongs to
        """
        self.node_id = node_id
        self.dev_id = dev_id
        self.qdisc_list = []
        self.class_list = []
        self.filter_list = []

    def add_qdisc(self, qdisc, parent="root", handle="", **kwargs):
        """
        Add a qdisc (Queueing Discipline) to this device

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        dev : Interface class
            The device to which the qdisc is to be added
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        handle : string
            id of the filter (Default value = '')
        """
        self.qdisc_list.append(
            Qdisc(self.node_id, self.dev_id, qdisc, parent, handle, **kwargs)
        )

        # Add qdisc to TopologyMap
        TopologyMap.add_qdisc(self.node_id, self.dev_id, qdisc, handle, parent=parent)

    def change_qdisc(self, handle, qdisc="", **kwargs):
        """
        Change a qdisc that is already present in the device

        Parameters
        ----------
        handle : string
            Hande of the qdisc to be changed
        qdisc : string
            The new qdisc to be added to the device
        """

        for qdisc_member in self.qdisc_list:
            if qdisc_member.handle == handle:
                # In case you want to just change kwargs
                if qdisc == "":
                    qdisc = qdisc_member.qdisc
                else:
                    TopologyMap.change_qdisc(self.node_id, self.dev_id, qdisc, handle)
                    qdisc_member.qdisc = qdisc
                engine.change_qdisc(
                    self.node_id,
                    self.dev_id,
                    qdisc,
                    qdisc_member.parent,
                    handle,
                    **kwargs,
                )

    def delete_qdisc(self, handle):
        """
        Delete qdisc (Queueing Discipline) from this device

        Parameters
        ----------
        handle : string
            Handle of the qdisc to be deleted
        """
        # TODO: Handle this better by using the destructor in traffic-control
        counter = 0
        for qdisc in self.qdisc_list:
            if qdisc.handle == handle:
                engine.delete_qdisc(
                    qdisc.node_id, qdisc.dev_id, qdisc.parent, qdisc.handle
                )
                TopologyMap.delete_qdisc(self.node_id, self.dev_id, handle)
                del self.qdisc_list[counter]
                break
            counter += 1

    def add_class(self, qdisc, parent="root", classid="", **kwargs):
        """
        Create an object that represents a class

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        classid : string
            id of the class (Default value = '')
        """
        self.class_list.append(
            Class(self.node_id, self.dev_id, qdisc, parent, classid, **kwargs)
        )

    def change_class(self, qdisc, parent, classid, **kwargs):
        """
        Change a class that is already present in the device

        Parameters
        ----------
        qdisc : string
            The qdisc which needs to be added to the device
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        classid : string
            id of the class (Default value = '')
        """
        for tc_class in self.class_list:
            if tc_class.classid == classid:
                tc_class.qdisc = qdisc
                tc_class.parent = parent
                engine.change_class(
                    self.node_id, self.dev_id, parent, qdisc, classid, **kwargs
                )

    def delete_class(self, classid, parent):
        """
        Delete Class from this device

        Parameters
        ----------
        classid : string
            Class id of the class to be deleted
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        """
        counter = 0
        for tc_class in self.class_list:  # tc_class since class is a keyword
            if tc_class.classid == classid:
                engine.delete_class(
                    self.node_id,
                    self.dev_id,
                    parent,
                    classid,
                )
                del self.class_list[counter]
                break
            counter += 1

    # pylint: disable=too-many-arguments
    def add_filter(
        self,
        priority,
        filtertype,
        flowid,
        protocol="ip",
        parent="root",
        handle="",
        **kwargs,
    ):
        """
        Design a Filter to assign to a Class or Qdisc

        Parameters
        ----------
        protocol : string
            protocol used (Default value = 'ip')
        priority : int
            priority of the filter
        filtertype : string
            one of the available filters
        flowid : Class
            classid of the class where the traffic is enqueued
            if the traffic passes the filter
        parent : string
            id of the parent class in major:minor form(optional) (Default value = 'root')
        handle : string
            id of the filter (Default value = '')
        filter : dictionary
            filter parameters
        """
        # TODO: Verify type of parameters
        # TODO: Reduce arguments to the engine functions by finding parent and handle automatically

        self.filter_list.append(
            Filter(
                self.node_id,
                self.dev_id,
                protocol,
                priority,
                filtertype,
                flowid,
                parent,
                handle,
                **kwargs,
            )
        )

    def delete_filter(self, handle, parent):
        """
        Delete filter from a Class or a Qdisc

        Parameters
        ----------
        handle : string
            Handle of the class to be deleted
        """
        counter = 0
        for tc_filter in self.filter_list:  # tc_class since class is a keyword
            if tc_filter.handle == handle:
                engine.delete_filter(
                    self.node_id,
                    self.dev_id,
                    parent,
                    handle,
                )
                del self.filter_list[counter]
                break
            counter += 1

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}(node_id = {self.node_id!r}, dev_id = {self.dev_id!r})"


class Qdisc:
    """Handle Queueing Discipline"""

    def __init__(self, node_id, dev_id, qdisc, parent="root", handle="", **kwargs):
        """
        Constructor to add a qdisc (Queueing Discipline) to an interface (device)

        Parameters
        ----------
        node_id : str
            The id of the namespace to which the interface belongs to
        dev_id : str
            The id of the interface to which the qdisc is to be added
        qdisc : str
            The qdisc which needs to be added to the interface
        dev_id : str
            The id of the interface to which the qdisc is to be added
        node_id : str
            The id of the namespace that the device belongs to
        parent : str
            id of the parent class in major:minor form(optional)
        handle : str
            id of the filter
        """
        self.node_id = node_id
        self.dev_id = dev_id
        self.qdisc = qdisc
        self.parent = parent
        self.handle = handle

        engine.add_qdisc(node_id, dev_id, qdisc, parent, handle, **kwargs)


class Class:
    """Handle classes associated to qdisc"""

    def __init__(self, node_id, dev_id, qdisc, parent="root", classid="", **kwargs):
        """
        Constructor to create an object that represents a class

        Parameters
        ----------
        node_id : str
            The id of the namespace to which the interface belongs to
        dev_id : str
            The id of the interface to which the qdisc is to be added
        node_id : str
            The id of the namespace that the device belongs to
        qdisc : str
            The qdisc which needs to be added to the interface
        parent : str
            id of the parent class in major:minor form(optional)
        classid : str
            id of the class
        """
        self.node_id = node_id
        self.dev_id = dev_id
        self.qdisc = qdisc  # NOTE: should be renamed to knid
        self.parent = parent
        self.classid = classid

        engine.add_class(node_id, dev_id, parent, qdisc, classid, **kwargs)


class Filter:
    """Handle filters to assign to class/qdisc"""

    def __init__(
        self,
        node_id,
        dev_id,
        protocol,
        priority,
        filtertype,
        flowid,
        parent="root",
        handle="",
        **kwargs,
    ):
        """
        Constructor to design a Filter to assign to a Class or Qdisc

        Parameters
        ----------
        node_id : str
            The id of the namespace to which the interface belongs to
        dev_id : str
            The id of the interface to which the qdisc is to be added
        protocol : str
            protocol used
        priority : int
            priority of the filter
        filtertype : str
            one of the available filters
        flowid : Class
            classid of the class where the traffic is enqueued
            if the traffic passes the filter
        parent : str
            id of the parent class in major:minor form(optional)
        handle : str
            id of the filter
        filter : dictionary
            filter parameters
        """
        self.node_id = node_id
        self.dev_id = dev_id
        self.protocol = protocol
        self.priority = priority
        self.filtertype = filtertype
        self.flowid = flowid
        self.parent = parent
        self.handle = handle

        engine.add_filter(
            node_id, dev_id, protocol, priority, filtertype, parent, handle, **kwargs
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Engine sub-package provides low-level APIs to other sub-packages

All the calls to Linux kernel happens in this module.
"""

from .exec import *
from .ip_address import *
from .ip_link import *
from .ip_mpls_route import *
from .ip_netns import *
from .ip_route import *
from .iperf3 import *
from .ping import *
from .setns import *
from .sysctl import *
from .tc import *
from .uname import *
from .ethtool import *



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test topology map API"""

import unittest
from nest.topology_map import TopologyMap

# pylint: disable=missing-docstring
# pylint: disable=too-many-instance-attributes


class TestTopologyMap(unittest.TestCase):
    def setUp(self):
        # Define namespace id and names
        self.ns_id1 = "ns_id1"
        self.ns_id2 = "ns_id2"
        self.ns_name1 = "ns_name1"
        self.ns_name2 = "ns_name2"

        # Add namespaces
        TopologyMap.add_namespace(self.ns_id1, self.ns_name1)
        TopologyMap.add_namespace(self.ns_id2, self.ns_name2)

        # Define interface id and names
        self.int_id1 = "int_id1"
        self.int_id2 = "int_id2"
        self.int_name1 = "int_name1"
        self.int_name2 = "int_name2"

        # Add interfaces
        TopologyMap.add_interface(self.ns_id1, self.int_id1, self.int_name1)
        TopologyMap.add_interface(self.ns_id1, self.int_id2, self.int_name2)

    def tearDown(self):
        TopologyMap.delete_all_mapping()

    def test_add_and_get_namespace(self):
        self.assertEqual(TopologyMap.get_namespace(self.ns_id1)["name"], self.ns_name1)
        self.assertEqual(TopologyMap.get_namespace(self.ns_id2)["name"], self.ns_name2)

    def test_add_and_get_interface(self):
        self.assertEqual(
            TopologyMap.get_interface(self.ns_id1, self.int_id1)["name"], self.int_name1
        )
        self.assertEqual(
            TopologyMap.get_interface(self.ns_id1, self.int_id2)["name"], self.int_name2
        )

    def test_add_same_entity_again(self):
        with self.assertRaises(ValueError):
            TopologyMap.add_namespace(self.ns_id1, self.ns_name1)
        with self.assertRaises(ValueError):
            TopologyMap.add_interface(self.ns_id1, self.int_id1, self.int_name1)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Class to handle zebra"""

from functools import partial
from nest.exceptions import RequiredDependencyNotFound
from nest.routing.route_daemons import RoutingDaemonBase
from nest.engine.dynamic_routing import run_zebra


class Zebra(RoutingDaemonBase):
    """
    Handles zebra related functionalities.
    Refer to `DaemonBase` for usage
    """

    def __init__(self, router_ns_id, interfaces, conf_dir, **kwargs):
        super().__init__(router_ns_id, interfaces, "zebra", conf_dir, **kwargs)

    def add_interface(self, interface):
        """
        Add interface command to config file
        """
        self.add_to_config(f"interface {interface}")

    def add_ip_address(self, ip_address):
        """
        Add IP address command to config file
        """
        if self.ipv6:
            self.add_to_config(f" ipv6 address {ip_address}")
        else:
            self.add_to_config(f" ip address {ip_address}")

    def create_basic_config(self):
        """
        Creates a file with basic configuration for OSPF.
        Use base `add_to_config` directly for more complex configurations
        """

        # Add loopback interface
        self.add_interface("lo")
        self.add_to_config(" no shutdown")
        for interface in self.interfaces:
            self.add_interface(interface.id)
            self.add_ip_address(interface.address.get_addr())
        if self.log_file is not None:
            self.add_to_config(f"log file {self.log_file}")

    def run(self):
        """
        Runs the zebra daemon
        """
        super().run(
            engine_func=partial(
                run_zebra, self.router_ns_id, self.conf_file, self.pid_file
            )
        )

    def handle_dependecy_error(self):
        self.logger.error(
            "Zebra not found. Routes from routing protocols cannot be added."
        )
        raise RequiredDependencyNotFound()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Execute Linux commands"""

import logging
import shlex
import subprocess
from subprocess import Popen, PIPE


logger = logging.getLogger(__name__)

# pylint: disable=inconsistent-return-statements
def exec_subprocess(cmd, shell=False, output=False):
    """
    Executes a command

    Parameters
    ----------
    cmd : str
        command to be executed
    shell : boolean
        Spawns a shell and executes the command if true
        (Default value = False)
    output : boolean
        True if the output of the `cmd` is to be returned
        (Default value = False)

    Returns
    -------
    int
        Return code received after executing the command
    """

    # TODO: Commands with pipes are easily executed when
    # they are run within a shell.
    # But it may not be the most efficient way.

    temp_cmd = cmd
    if shell is False:
        temp_cmd = cmd.split()

    with Popen(temp_cmd, stdout=PIPE, stderr=PIPE, shell=shell) as proc:

        (stdout, _) = proc.communicate()
        logger.trace(cmd)

        if output:
            return stdout.decode()
        return proc.returncode


def exec_subprocess_with_live_output(cmd):
    """
    Executes command and prints live output to stdout.

    For eg., if ping command is used, then the output for each
    packet is printed live to stdout.

    Parameters
    ----------
    cmd : str
        Command to be executed

    Returns
    -------
    int
        Return code recieved after executing the command
    """
    # Inspired from:
    # https://fabianlee.org/2019/09/15/python-getting-live-output-from-subprocess-using-poll/

    with Popen(cmd.split(), stdout=PIPE) as process:
        while True:
            output = process.stdout.readline()
            if process.poll() is not None:
                break
            if output:
                print(output.decode(), end="")

        # Print an extra newline at the end
        print()
        logger.trace(cmd)
        return process.poll()


def exec_exp_commands(cmd, stdout=PIPE, stderr=PIPE, timeout=None):
    """
    executes experiment related commands like ss, tc and netperf

    Parameters
    ----------
    cmd : str
        command to be executed
    stdout : File
        temp file(usually) to store the output (Default value = subprocess.PIPE)
    stderr : File
        temp file(usually) to store errors, if any (Default value = subprocess.PIPE)
    timeout :
         (Default value = None)

    Returns
    -------
    int
        Return code recieved after executing the command
    """
    with Popen(shlex.split(cmd), stdout=stdout, stderr=stderr) as proc:
        logger.trace(cmd)
        try:
            proc.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            proc.kill()
            stderr.write(b"Connection timeout")
        return proc.returncode



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""iperf commands"""

from .exec import exec_subprocess, exec_exp_commands


def run_iperf_server(ns_name):
    """
    Run Iperf Server on a namesapce

    Parameters
    ----------
    ns_name : str
        name of the server namespace
    """
    # Runs server as a daemon
    return exec_subprocess(f"ip netns exec {ns_name} iperf3 -s -D")


# pylint: disable=too-many-arguments
def run_iperf_client(ns_name, iperf3_options, destination_ip, ipv6, out, err):
    """
    Run Iperf Client

    Parameters
    ----------
    ns_name : str
        name of the client namespace
    iperf3_options : str
        Specific options (like run_time, number_of_flows) to run iperf3 command with
    destination_ip : str
        the ip of server to which it has to connect
    ipv6 : bool
        determines if destination_ip is ipv4/ipv6
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors

    Returns
    -------
    int
        return code of the command executed
    """
    if ipv6:
        return exec_exp_commands(
            f"ip netns exec {ns_name} iperf3 -6 -u -c {destination_ip} {iperf3_options}",
            stdout=out,
            stderr=err,
        )

    return exec_exp_commands(
        f"ip netns exec {ns_name} iperf3 -u -c {destination_ip} {iperf3_options}",
        stdout=out,
        stderr=err,
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""
Run commands inside a network namespace

Inspired from: https://github.com/larsks/python-netns
"""

import os
from ctypes import CDLL, get_errno

CLONE_NEWNET = 0x40000000


def _errcheck(ret, _func, _args):
    if ret == -1:
        err = get_errno()
        raise OSError(err, os.strerror(err))


libc = CDLL("libc.so.6", use_errno=True)
libc.setns.errcheck = _errcheck


def get_ns_path(ns_name):
    """
    Get path to network namespace.

    Parameters
    ----------
    ns_name : str
        namespace name
    """
    ns_path = f"/var/run/netns/{ns_name}"

    if not os.path.exists(ns_path):
        raise ValueError(f"Network namespace path {ns_path} does not exist")

    return ns_path


def set_ns(ns_name=None):
    """
    Change the network namespace of the calling thread.

    If ns_name is None, then default namespace is assumed.

    Parameters
    ----------
    ns_name : str/None
        namespace name
    """
    ns_path = None

    if ns_name is None:
        # Default namespace, set ns_path to init process network namespace
        ns_path = "/proc/1/ns/net"
    else:
        ns_path = get_ns_path(ns_name)

    with open(ns_path) as file:
        fd = file.fileno()  # pylint: disable=invalid-name
        libc.setns(fd, CLONE_NEWNET)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""Test APIs from input_validator sub-package"""

import unittest
from typing import List
from nest.input_validator import input_validator
from nest.topology import Node, Address

# pylint: disable=missing-docstring


@input_validator
def samplemethod(arg1: int, arg2: str):
    assert isinstance(arg1, int)
    assert isinstance(arg2, str)


@input_validator
def samplemethod2(node: Node, address: Address):
    assert isinstance(node, Node)
    assert isinstance(address, Address)


@input_validator
def samplemethod3(interfaces: List[Node], address: Address = None):
    assert isinstance(interfaces, list)
    assert address is None


class TestInputValidator(unittest.TestCase):
    def test_expected_type(self):
        samplemethod(1, "2")

        with self.assertRaises(TypeError) as ex:
            samplemethod("1", 2)

        self.assertEqual(
            str(ex.exception),
            "Expected type of argument 'arg1' in method 'samplemethod' "
            "is <class 'int'>. But got input '1' of type <class 'str'>",
        )

        with self.assertRaises(TypeError) as ex:
            samplemethod(1, 2)

        self.assertEqual(
            str(ex.exception),
            "Expected type of argument 'arg2' in method 'samplemethod' "
            "is <class 'str'>. But got input '2' of type <class 'int'>",
        )

    def test_type_casting(self):
        samplemethod2(Node("n1"), Address("10.0.0.1/24"))

        samplemethod2(Node("n2"), "10.0.0.1/24")

        with self.assertRaises(TypeError) as ex:
            samplemethod2("n3", "10.0.0.1/24")

        self.assertEqual(
            str(ex.exception),
            "Expected type of argument 'node' in method 'samplemethod2' "
            "is <class 'nest.topology.node.Node'>. But got input 'n3' of type <class 'str'>",
        )

        with self.assertRaises(TypeError) as ex:
            samplemethod2(Node("n4"), "10.0.0.1/244")

        self.assertEqual(
            str(ex.exception),
            "For argument 'address' in method 'samplemethod2', "
            "converting '10.0.0.1/244' to type Address failed.\n"
            "Please see the previous exception to know why the conversion failed.",
        )

    def test_optional_and_list_field(self):
        samplemethod3([Node("h1"), Node("h2")])

        with self.assertRaises(TypeError) as ex:
            samplemethod3(Node("h3"))

        self.assertEqual(
            str(ex.exception),
            "Expected type of argument 'interfaces' in method 'samplemethod3' "
            "is List. But got input 'Node('h3')' of type <class 'nest.topology.node.Node'>",
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

""" Util functions """

from .exec import exec_subprocess


def is_dependency_installed(tool):
    """
    util to check if a tool is installed

    Parameters
    ----------
    tool : str
        tool to check

    Returns
    -------
    bool
        true if the `tool` is installed

    """
    return_code = exec_subprocess(f"type {tool}", shell=True)
    return return_code == 0



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to router creation in topology"""

from nest.topology.node import Node


class Router(Node):
    """
    Abstraction for a Router.

    Attributes
    ----------
    name: str
        User given name for the Router.
    """

    def __init__(self, name):
        """
        Create a router with given `name`.

        Parameters
        ----------
        name: str
            The name of the router to be created
        """
        super().__init__(name)
        self.enable_ip_forwarding()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Base class for other runners
"""

import tempfile
import logging
from nest.topology import Address
from nest.topology_map import TopologyMap


# pylint: disable=too-many-instance-attributes
class Runner:
    """
    Base class for other runners

    Attributes
    ----------
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors
    start_time : num
            time at which utility is to run
    run_time : num
        total time for a utility to run
    destination_address : Address
        Address of the destination node for the runner
    """

    # pylint: disable=too-many-arguments
    def __init__(self, ns_id, start_time, run_time, destination_ip="::1", dst_ns=None):
        """
        Parameters
        ----------
        ns_id: str
            Namespace where the utility is run
        destination_ip: str
            ip address of the destination namespace
        dst_ns: str
            Optional. The namespace where the utility either sends traffic to,
            or monitors traffic that is going from `ns_id` to `dst_ns`
        """
        # pylint: disable=consider-using-with
        self.out = tempfile.TemporaryFile()
        self.err = tempfile.TemporaryFile()

        self.logger = logging.getLogger(__name__)

        self.ns_id = ns_id
        self.dst_ns = dst_ns
        self.start_time = start_time
        self.run_time = run_time
        self.destination_address = Address(destination_ip)

    def run(self, engine_func, error_string_prefix="Error"):
        """
        executes the given engine function and prints error(if any)

        Parameters
        ----------
        engine_func: Function
            engine function to be called
        """
        try:
            return_code = engine_func(out=self.out, err=self.err)
            if return_code != 0:
                self.print_error(error_string_prefix)
        except KeyboardInterrupt:
            ns_name = TopologyMap.get_namespace(self.ns_id)["name"]
            self.logger.debug(
                "%s at %s: Received KeyboardInterrupt, hence shutting down the process gracefully.",
                error_string_prefix,
                ns_name,
            )

    def print_error(self, error_string_prefix):
        """
        Method to print error from `self.err`
        """
        self.err.seek(0)  # rewind to start of file
        error = self.err.read().decode()
        ns_name = TopologyMap.get_namespace(self.ns_id)["name"]
        self.logger.error("%s at %s. %s", error_string_prefix, ns_name, error)

    def get_meta_item(self):
        """
        Return the meta item for the given flow.
        This "meta" information is required by plotter.
        """
        meta_item = {
            "meta": True,
            "start_time": str(self.start_time),
            "stop_time": str(self.start_time + self.run_time),
        }

        if self.dst_ns is not None:
            meta_item["destination_node"] = TopologyMap.get_namespace(self.dst_ns)[
                "name"
            ]

        return meta_item

    def __del__(self):
        """
        Close the temp files created
        """
        self.out.close()
        self.err.close()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

import pathlib
import sys
from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()

# Get version
version = {}
with open("nest/version.py") as fp:
    exec(fp.read(), version)

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

# Check for pip versions below 9.0.0
if sys.version_info < (3, 6) or sys.version_info >= (4, 0):
    raise RuntimeError("This package requires Python version >=3.6 and <4")

setup(
    name="nest",
    version=version["__version__"],
    description="NeST: Network Stack Tester",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://nest.nitk.ac.in/",
    author="NITK",
    author_email="nest@nitk.edu.in",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Topic :: Education",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="network, namespace, linux",
    packages=find_packages(),
    package_data={
        "nest.experiment": ["info/README.txt"],
        "nest.experiment.parser": ["iterators/*.sh"],
        "nest": ["config.json"],
    },
    python_requires=">=3.6, <4",
    install_requires=["matplotlib", "numpy", "packaging", "tqdm"],
    project_urls={
        "Source": "https://gitlab.com/nitk-nest/nest",
    },
)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from topology sub-package"""

import unittest
from nest.topology import Node, connect
from nest.experiment import Experiment, Flow
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap
import nest.config as config

# pylint: disable=missing-docstring
# pylint: disable=invalid-name
class TestExperiment(unittest.TestCase):
    def test_experiment(self):
        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms", "pie")
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment")
        flow = Flow(n0, n1, n1_r.address, 0, 5, 2)
        exp.add_tcp_flow(flow)

        exp.run()

    def test_experiment_udp(self):
        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms", "pie")
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment-udp")
        flow = Flow(n0, n1, n1_r.address, 0, 5, 2)
        exp.add_udp_flow(flow)

        exp.run()

    def test_experiment_ipv6(self):
        # Test IPv6 with Duplicate Address Detection (DAD) disabled
        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10::1:1/122")
        r_n0.set_address("10::1:2/122")
        r_n1.set_address("10::2:2/122")
        n1_r.set_address("10::2:1/122")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms", "pie")
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment-ipv6")
        flow = Flow(n0, n1, n1_r.address, 0, 5, 2)
        exp.add_tcp_flow(flow)

        exp.run()

    def test_experiment_ipv6_dad(self):
        # Test IPv6 with Duplicate Address Detection (DAD) enabled
        config.set_value("disable_dad", False)

        n0 = Node("n0")
        n1 = Node("n1")
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(n0, r)
        (r_n1, n1_r) = connect(r, n1)

        n0_r.set_address("10::1:1/122")
        r_n0.set_address("10::1:2/122")
        r_n1.set_address("10::2:2/122")
        n1_r.set_address("10::2:1/122")

        n0.add_route("DEFAULT", n0_r)
        n1.add_route("DEFAULT", n1_r)

        n0_r.set_attributes("100mbit", "5ms")
        r_n0.set_attributes("100mbit", "5ms")

        r_n1.set_attributes("10mbit", "40ms", "pie")
        n1_r.set_attributes("10mbit", "40ms")

        exp = Experiment("test-experiment-ipv6-dad")
        flow = Flow(n0, n1, n1_r.address, 0, 5, 2)
        exp.add_tcp_flow(flow)

        exp.run()

        # Resetting disable_dad in config
        config.set_value("disable_dad", True)

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal
"""
Custom exceptions
"""


class RequiredDependencyNotFound(Exception):
    """
    Raised when a depedency required for a experiment/topology
    is not installed.
    """

    def __init__(self, message="Depedency required for this program is not installed"):
        self.message = message
        super().__init__(self.message)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Plot ss results"""

import logging
import matplotlib.pyplot as plt
from nest.experiment.interrupts import handle_keyboard_interrupt
from ..pack import Pack
from .common import simple_plot, mix_plot

logger = logging.getLogger(__name__)


def _get_list_of_ss_params():
    """
    Return list of params parsed by ss
    """
    return ["cwnd", "rtt", "dev_rtt", "ssthresh", "rto", "delivery_rate", "pacing_rate"]


def _get_ss_params_units():
    """
    Obtain the units of each ss parameters
    """
    units = {
        "cwnd": "Packets",
        "rtt": "ms",
        "dev_rtt": "ms",
        "ssthresh": "Packets",
        "rto": "ms",
        "delivery_rate": "Mbps",
        "pacing_rate": "Mbps",
    }
    return units


def _get_nicer_param_names():
    """
    For example,
    'rtt' -> 'RTT'
    'delivery_rate' -> 'Delivery rate'
    """
    nice_names = {
        "cwnd": "Congestion Window",
        "rtt": "RTT",
        "dev_rtt": "Deviation in RTT",
        "ssthresh": "Slow Start Threshold",
        "rto": "TCP Retransmission Timeout",
        "delivery_rate": "Delivery Rate",
        "pacing_rate": "Pacing Rate",
    }
    return nice_names


def _get_ylabel(param):
    """
    Obtain y label for plots for a given parameter.

    For example, 'delivery_rate' -> 'Delivery rate (Mbps)'

    Arguments
    ---------
        param : str
            parameter

    Returns
    -------
        str
    """
    # Get ss param units
    units = _get_ss_params_units()
    # Get "nicer" param names
    nicer_param_names = _get_nicer_param_names()

    ylabel = nicer_param_names[param]
    if units[param] is not None:
        ylabel += f" ({units[param]})"

    return ylabel


def _extract_from_ss_flow(flow, node, dest_ip, dest_port):
    """
    Extract information from flow data and convert it to
    conveniently plottable data format

    Parameters
    ----------
    flow : List
        List with timestamps and stats
    node : string
        Node from which ss results were obtained from
    dest_ip : string
        Destination ip address of the flow
    dest_port : string
        Destination port of the flow

    Returns
    -------
    (timestamp, flow_params) or None
        Return timestamp and flow parameters
    """
    # "meta" item will always be present, hence `<= 1`
    if len(flow) <= 1:
        logger.warning(
            "Flow from %s to destination %s:%s " "doesn't have any parsed ss result.",
            node,
            dest_ip,
            dest_port,
        )
        return None

    # First item is the "meta" item with user given information
    user_given_start_time = float(flow[0]["start_time"])
    destination_node = flow[0]["destination_node"]

    # "Bias" actual start_time in experiment with user given start time
    start_time = float(flow[1]["timestamp"]) - user_given_start_time

    timestamp = []
    flow_params = {}

    for param in _get_list_of_ss_params():
        flow_params[param] = []

    for data in flow[1:]:
        for stat in flow_params:
            if stat in data:
                flow_params[stat].append(float(data[stat]))
            else:
                flow_params[stat].append(None)
        relative_time = float(data["timestamp"]) - start_time
        timestamp.append(relative_time)

    return {"destination_node": destination_node, "values": (timestamp, flow_params)}


def _plot_ss_flow(flow, node, dest_ip, dest_port):
    """
    Plot ss stats of the flow

    Parameters
    ----------
    exp_name : string
        Name of experiment for which results were obtained
    flow : List
        List with timestamps and stats
    node : string
        Node from which ss results were obtained from
    dest_ip : string
        Destination ip address of the flow
    dest_port : string
        Destination port of the flow
    """
    data = _extract_from_ss_flow(flow, node, dest_ip, dest_port)
    destination_node = data["destination_node"]
    values = data["values"]

    if values is None:
        return None
    (timestamp, flow_params) = values

    legend_string = f"{node} to {destination_node} ({dest_ip}:{dest_port})"
    for param in flow_params:
        fig = simple_plot(
            "Socket Statistics",
            timestamp,
            flow_params[param],
            "Time (Seconds)",
            _get_ylabel(param),
            legend_string=legend_string,
        )

        filename = f"{param}_{node}_to_{destination_node}({dest_ip}:{dest_port}).png"
        Pack.dump_plot("ss", filename, fig)
        plt.close(fig)

    return {
        "label": legend_string,
        "destination_node": destination_node,
        "values": (timestamp, flow_params),
    }


# pylint: disable=too-many-locals
@handle_keyboard_interrupt
def plot_ss(parsed_data):
    """
    Plot statistics obtained from ss

    Parameters
    ----------
    exp_name : string
        Name of experiment for which results were obtained
    parsed_data : Dict
        JSON data parsed from ss
    """

    # pylint: disable=too-many-nested-blocks
    for node in parsed_data:
        node_data = parsed_data[node]
        for connection in node_data:
            for dest_ip in connection:

                all_flow_data = []
                destination_node = None

                flow_data = connection[dest_ip]
                for dest_port in flow_data:
                    flow = flow_data[dest_port]
                    plotted_data = _plot_ss_flow(flow, node, dest_ip, dest_port)
                    if plotted_data is not None:
                        all_flow_data.append(
                            {
                                "label": plotted_data["label"],
                                "values": plotted_data["values"],
                            }
                        )

                        if destination_node is None:
                            destination_node = plotted_data["destination_node"]
                        elif destination_node != plotted_data["destination_node"]:
                            raise Exception(
                                "Error in plotting ss stats. Unexpected "
                                "destination node in ss.json"
                            )

                if len(all_flow_data) > 1:

                    x_vals = []
                    labels = []

                    for flow_data in all_flow_data:
                        x_vals.append(flow_data["values"][0])
                        labels.append(flow_data["label"])

                    for param in all_flow_data[0]["values"][1]:
                        y_vals = []
                        for flow_data in all_flow_data:
                            y_vals.append(flow_data["values"][1][param])

                        data = []

                        for i in range(len(labels)):
                            data.append(
                                {"values": (x_vals[i], y_vals[i]), "label": labels[i]}
                            )

                        fig = mix_plot(
                            "Socket Statistics",
                            data,
                            "Time (Seconds)",
                            _get_ylabel(param),
                        )
                        filename = (
                            f"{param}_{node}_to_{destination_node}({dest_ip}).png"
                        )
                        Pack.dump_plot("ss", filename, fig)
                        plt.close(fig)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal
"""Test APIs from topology sub-package"""

import unittest
import subprocess
from nest.topology_map import TopologyMap
from nest.topology import Node, connect
from nest.clean_up import delete_namespaces

# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=consider-using-with


class TestOffloads(unittest.TestCase):
    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

    def test_enable_offload(self):
        n2 = Node("n2")
        n1 = Node("n1")

        (n1_n2, n2_n1) = connect(n1, n2)

        # Enable TSO offload on n1_n2 interface
        n1_n2.enable_offload("tso")
        with n1:
            proc = subprocess.Popen(
                f"ip netns exec {n1.id}  ethtool -k {n1_n2.id}",
                shell=True,
                stdout=subprocess.PIPE,
            )
            (output, _) = proc.communicate()
            self.assertTrue("tcp-segmentation-offload: on" in output.decode())

        # Enable offloads GSO, GRO on n2_n1 interface
        n2_n1.enable_offload(["gso", "gro"])
        with n2:
            proc = subprocess.Popen(
                f"ip netns exec {n2.id}  ethtool -k {n2_n1.id}",
                shell=True,
                stdout=subprocess.PIPE,
            )
            (output, _) = proc.communicate()
            self.assertTrue("generic-segmentation-offload: on" in output.decode())
            self.assertTrue("generic-receive-offload: on" in output.decode())

    def test_disable_offload(self):
        n2 = Node("n2")
        n1 = Node("n1")

        (n1_n2, n2_n1) = connect(n1, n2)

        # Disable TSO offload on n1_n2 interface
        n1_n2.disable_offload("tso")
        with n1:
            proc = subprocess.Popen(
                f"ip netns exec {n1.id}  ethtool -k {n1_n2.id}",
                shell=True,
                stdout=subprocess.PIPE,
            )
            (output, _) = proc.communicate()
            self.assertTrue("tcp-segmentation-offload: off" in output.decode())

        # Disable offloads GSO, GRO on n2_n1 interface
        n2_n1.disable_offload(["gso", "gro"])
        with n2:
            proc = subprocess.Popen(
                f"ip netns exec {n2.id}  ethtool -k {n2_n1.id}",
                shell=True,
                stdout=subprocess.PIPE,
            )
            (output, _) = proc.communicate()
            self.assertTrue("generic-segmentation-offload: off" in output.decode())
            self.assertTrue("generic-receive-offload: off" in output.decode())


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Generate unique id for topology entity"""

from .. import config


class IdGen:  # pylint: disable=too-few-public-methods
    """Generate unique id for Topology entities


    Attributes
    ----------
    topology_id: str
        Id prefixed for each topology entity
    counter: int
        Unique identifier of an entity within a topology
    is_unique_id: bool
        If unique Id's should be used for entity names
        (default is True)

    """

    topology_id = ""
    counter = 0

    def __init__(self, topology_id):
        """Initialize `topology_id`

        Parameters
        ----------
        topology_id: str

        """
        IdGen.topology_id = topology_id

    @staticmethod
    def get_id(name):
        """Generate unique id on each call

        Parameters
        ----------
        name: str
            Name of the entity as shown to the user

        Returns
        -------
        str
            If `is_unique_id` is true, then an unique id is returned
            Else, `name` is returned back

        """
        if config.get_value("assign_random_names"):
            IdGen.counter += 1
            return IdGen.topology_id + "-" + str(IdGen.counter)

        return name



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Script to be run for running experiments on topology"""

from multiprocessing import Process
from collections import namedtuple, defaultdict
import logging
import os
from time import sleep
from tqdm import tqdm

from nest.logging_helper import DepedencyCheckFilter
from nest import config
from nest.topology_map import TopologyMap
from nest.clean_up import kill_processes
from .pack import Pack

# Import results
from .results import SsResults, NetperfResults, Iperf3Results, TcResults, PingResults

# Import parsers
from .parser.ss import SsRunner
from .parser.netperf import NetperfRunner
from .parser.iperf3 import Iperf3Runner
from .parser.tc import TcRunner
from .parser.ping import PingRunner

# Import plotters
from .plotter.ss import plot_ss
from .plotter.netperf import plot_netperf
from .plotter.iperf3 import plot_iperf3
from .plotter.tc import plot_tc
from .plotter.ping import plot_ping
from ..engine.util import is_dependency_installed

logger = logging.getLogger(__name__)
if not any(isinstance(filter, DepedencyCheckFilter) for filter in logger.filters):
    # Duplicate filter is added to avoid logging of same error
    # messages incase any of the tools is not installed
    logger.addFilter(DepedencyCheckFilter())

# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def run_experiment(exp):
    """
    Run experiment

    Parameters
    -----------
    exp : Experiment
        The experiment attributes
    """

    tools = ["netperf", "ss", "tc", "iperf3", "ping"]
    Runners = namedtuple("runners", tools)
    exp_runners = Runners(
        netperf=[], ss=[], tc=[], iperf3=[], ping=[]
    )  # Runner objects

    # Keep track of all destination nodes [to ensure netperf and iperf3
    # server is run at most once]
    destination_nodes = {"netperf": set(), "iperf3": set()}

    # Contains start time and end time to run respective command
    # from a source netns to destination address (in destination netns)
    ss_schedules = defaultdict(lambda: (float("inf"), float("-inf")))
    ping_schedules = defaultdict(lambda: (float("inf"), float("-inf")))

    # Overall experiment stop time considering all flows
    exp_end_t = float("-inf")

    dependencies = get_dependency_status(tools)

    ss_required = False
    ss_filters = set()

    # Traffic generation
    for flow in exp.flows:
        # Get flow attributes
        [
            src_ns,
            dst_ns,
            dst_addr,
            start_t,
            stop_t,
            _,
            options,
        ] = flow._get_props()  # pylint: disable=protected-access

        exp_end_t = max(exp_end_t, stop_t)

        (min_start, max_stop) = ping_schedules[(src_ns, dst_ns, dst_addr)]
        ping_schedules[(src_ns, dst_ns, dst_addr)] = (
            min(min_start, start_t),
            max(max_stop, stop_t),
        )

        # Setup TCP/UDP flows
        if options["protocol"] == "TCP":
            # * Ignore netperf tcp control connections
            # * Destination port of netperf control connection is 12865
            # * We also have "sport" (source port) in the below condition since
            #   there can be another flow in the reverse direction whose control
            #   connection also we must ignore.
            ss_filters.add("sport != 12865 and dport != 12865")
            ss_required = True
            (tcp_runners, ss_schedules,) = setup_tcp_flows(
                dependencies["netperf"],
                flow,
                ss_schedules,
                destination_nodes["netperf"],
            )

            exp_runners.netperf.extend(tcp_runners)

            # Update destination nodes
            destination_nodes["netperf"].add(dst_ns)

        elif options["protocol"] == "UDP":
            # * Ignore iperf3 tcp control connections
            # * Destination port of iperf3  control connection is 5201
            # * We also have "sport" (source port) in the below condition since
            #   there can be another flow in the reverse direction whose control
            #   connection also we must ignore.
            ss_filters.add("sport != 5201 and dport != 5201")
            udp_runners = setup_udp_flows(
                dependencies["iperf3"], flow, destination_nodes["iperf3"]
            )

            exp_runners.iperf3.extend(udp_runners)

            # Update destination nodes
            destination_nodes["iperf3"].add(dst_ns)

    if ss_required:
        ss_filter = " and ".join(ss_filters)
        ss_runners = setup_ss_runners(dependencies["ss"], ss_schedules, ss_filter)
        exp_runners.ss.extend(ss_runners)

    tc_runners = setup_tc_runners(dependencies["tc"], exp.qdisc_stats, exp_end_t)
    exp_runners.tc.extend(tc_runners)

    ping_runners = setup_ping_runners(dependencies["ping"], ping_schedules)
    exp_runners.ping.extend(ping_runners)

    try:
        # Start traffic generation
        run_workers(setup_flow_workers(exp_runners, exp_end_t))

        logger.info("Parsing statistics...")

        # Parse the stored statistics
        run_workers(setup_parser_workers(exp_runners))

        logger.info("Parsing statistics complete!")
        logger.info("Output results as JSON dump...")

        # Output results as JSON dumps
        dump_json_ouputs()

        if config.get_value("readme_in_stats_folder"):
            # Copying README.txt to stats folder
            relative_path = os.path.join("info", "README.txt")
            readme_path = os.path.join(os.path.dirname(__file__), relative_path)
            Pack.copy_files(readme_path)

        if config.get_value("plot_results"):
            logger.info("Plotting results...")

            # Plot results and dump them as images
            run_workers(setup_plotter_workers())

            logger.info("Plotting complete!")

        logger.info("Experiment %s complete!", exp.name)

    except KeyboardInterrupt:
        logger.warning(
            "Experiment %s forcefully stopped. The results obtained maybe incomplete!",
            exp.name,
        )
    finally:
        cleanup()


def run_workers(workers):
    """
    Run and wait for processes to finish

    Parameters
    ----------
    workers: list[multiprocessing.Process]
        List of processes to be run
    """
    # Start workers
    for worker in workers:
        worker.start()

    # wait for all the workers to finish
    for worker in workers:
        worker.join()


def setup_plotter_workers():
    """
    Setup plotting processes

    Returns
    -------
    List[multiprocessing.Process]
        plotters
    """
    plotters = []

    plotters.append(Process(target=plot_ss, args=(SsResults.get_results(),)))
    plotters.append(Process(target=plot_netperf, args=(NetperfResults.get_results(),)))
    plotters.append(Process(target=plot_iperf3, args=(Iperf3Results.get_results(),)))
    plotters.append(Process(target=plot_tc, args=(TcResults.get_results(),)))
    plotters.append(Process(target=plot_ping, args=(PingResults.get_results(),)))

    return plotters


def dump_json_ouputs():
    """
    Outputs experiment results as json dumps
    """
    SsResults.output_to_file()
    NetperfResults.output_to_file()
    Iperf3Results.output_to_file()
    TcResults.output_to_file()
    PingResults.output_to_file()


def setup_flow_workers(exp_runners, exp_stop_time):
    """
    Setup flow generation and stats collection processes(netperf, ss, tc, iperf3...).

    Also add a progress bar process for showing experiment progress.

    Parameters
    ----------
    exp_runners: collections.NamedTuple
        all(netperf, ping, ss, tc..) the runners
    exp_stop_time: int
        Time when experiment stops (in seconds)

    Returns
    -------
    List[multiprocessing.Process]
        flow generation and stats collection processes
        + progress bar process
    """
    workers = []

    for runners in exp_runners:
        workers.extend([Process(target=runner.run) for runner in runners])

    # Add progress bar process
    if config.get_value("show_progress_bar"):
        workers.extend([Process(target=progress_bar, args=(exp_stop_time,))])

    return workers


def setup_parser_workers(exp_runners):
    """
    Setup parsing processes

    Parameters
    ----------
    exp_runners: collections.NamedTuple
        all(netperf, ping, ss, tc..) the runners

    Returns
    -------
    List[multiprocessing.Process]
        parsers
    """
    parsers = []

    for ss_runner in exp_runners.ss:
        parsers.append(Process(target=ss_runner.parse))

    for netperf_runner in exp_runners.netperf:
        parsers.append(Process(target=netperf_runner.parse))

    for iperf3_runner in exp_runners.iperf3:
        parsers.append(Process(target=iperf3_runner.parse))

    for tc_runner in exp_runners.tc:
        parsers.append(Process(target=tc_runner.parse))

    for ping_runner in exp_runners.ping:
        parsers.append(Process(target=ping_runner.parse))

    return parsers


def get_dependency_status(tools):
    """
    Checks for dependency

    Parameters
    ----------
    tools: List[str]
        list of tools to check for it's installation

    Returns
    -------
    dict
        contains information as to whether `tools` are installed
    """
    dependencies = {}
    for dependency in tools:
        dependencies[dependency] = is_dependency_installed(dependency)
    return dependencies


def setup_tcp_flows(dependency, flow, ss_schedules, destination_nodes):
    """
    Setup netperf to run tcp flows
    Parameters
    ----------
    dependency: int
        whether netperf is installed
    flow: Flow
        Flow parameters
    ss_schedules:
        ss_schedules so far
    destination_nodes:
        Destination nodes so far already running netperf server

    Returns
    -------
    dependency: int
        updated dependency in case netperf is not installed
    netperf_runners: List[NetperfRunner]
        all the netperf flows generated
    workers: List[multiprocessing.Process]
        Processes to run netperf flows
    ss_schedules: dict
        updated ss_schedules
    """
    netperf_runners = []
    if not dependency:
        logger.warning("Netperf not found. Tcp flows cannot be generated")
    else:
        # Get flow attributes
        [
            src_ns,
            dst_ns,
            dst_addr,
            start_t,
            stop_t,
            n_flows,
            options,
        ] = flow._get_props()  # pylint: disable=protected-access

        # Run netserver if not already run before on given dst_node
        if dst_ns not in destination_nodes:
            NetperfRunner.run_netserver(dst_ns)

        src_name = TopologyMap.get_namespace(src_ns)["name"]

        netperf_options = {}
        netperf_options["testname"] = "TCP_STREAM"
        netperf_options["cong_algo"] = options["cong_algo"]
        f_flow = "flow" if n_flows == 1 else "flows"
        logger.info(
            "Running %s netperf %s from %s to %s...",
            n_flows,
            f_flow,
            src_name,
            dst_addr,
        )

        # Create new processes to be run simultaneously
        for _ in range(n_flows):
            runner_obj = NetperfRunner(
                src_ns, dst_addr, start_t, stop_t - start_t, dst_ns, **netperf_options
            )
            netperf_runners.append(runner_obj)

        # Find the start time and stop time to run ss command in `src_ns` to a `dst_addr`
        ss_schedules = _get_start_stop_time_for_ss(
            src_ns, dst_ns, dst_addr, start_t, stop_t, ss_schedules
        )

    return netperf_runners, ss_schedules


def setup_udp_flows(dependency, flow, destination_nodes):
    """
    Setup iperf3 to run udp flows

    Parameters
    ----------
    dependency: int
        whether iperf3 is installed
    flow: Flow
        Flow parameters
    destination_nodes:
        Destination nodes so far already running iperf3 server

    Returns
    -------
    dependency: int
        updated dependency in case iproute2 is not installed
    iperf3_runners: List[NetperfRunner]
        all the iperf3 udp flows generated
    workers: List[multiprocessing.Process]
        Processes to run iperf3 udp flows
    """
    iperf3_runners = []
    if not dependency:
        logger.warning("Iperf3 not found. Udp flows cannot be generated")
    else:
        # Get flow attributes
        [
            src_ns,
            dst_ns,
            dst_addr,
            start_t,
            stop_t,
            n_flows,
            options,
        ] = flow._get_props()  # pylint: disable=protected-access

        # Run iperf3 server if not already run before on given dst_node
        if dst_ns not in destination_nodes:
            Iperf3Runner.run_server(dst_ns)

        src_name = TopologyMap.get_namespace(src_ns)["name"]
        f_flow = "flow" if n_flows == 1 else "flows"
        logger.info(
            "Running %s udp %s from %s to %s...", n_flows, f_flow, src_name, dst_addr
        )

        runner_obj = Iperf3Runner(
            src_ns,
            dst_addr,
            options["target_bw"],
            n_flows,
            start_t,
            stop_t - start_t,
            dst_ns,
        )
        iperf3_runners.append(runner_obj)

    return iperf3_runners


def setup_ss_runners(dependency, ss_schedules, ss_filter):
    """
    setup SsRunners for collecting tcp socket statistics

    Parameters
    ----------
    dependency: int
        whether ss is installed
    ss_schedules: dict
        start time and end time for SsRunners

    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run ss at nodes
    runners: List[SsRunners]
    """
    runners = []
    if dependency:
        logger.info("Running ss on nodes...")
        for key, timings in ss_schedules.items():
            src_ns = key[0]
            dst_ns = key[1]
            dst_addr = key[2]
            ss_runner = SsRunner(
                src_ns,
                dst_addr,
                timings[0],
                timings[1] - timings[0],
                dst_ns,
                ss_filter=ss_filter,
            )
            runners.append(ss_runner)
    else:
        logger.warning("ss not found. Sockets stats will not be collected")
    return runners


def setup_tc_runners(dependency, qdisc_stats, exp_end):
    """
    setup TcRunners for collecting qdisc statistics

    Parameters
    ----------
    dependency: int
        whether tc is installed
    qdisc_stats: dict
        info regarding nodes to run tc on
    exp_end: float
        time to stop running tc
    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run tc at nodes
    runners: List[TcRunners]
    """
    runners = []
    if dependency and len(qdisc_stats) > 0:
        logger.info("Running tc on requested interfaces...")
        for qdisc_stat in qdisc_stats:
            tc_runner = TcRunner(
                qdisc_stat["ns_id"], qdisc_stat["int_id"], qdisc_stat["qdisc"], exp_end
            )
            runners.append(tc_runner)
    elif not dependency:
        logger.warning("tc not found. Qdisc stats will not be collected")
    return runners


def setup_ping_runners(dependency, ping_schedules):
    """
    setup PingRunners for collecting latency

    Parameters
    ----------
    dependency: int
        whether ping is installed
    ping_schedules: dict
        start time and end time for PingRunners

    Returns
    -------
    workers: List[multiprocessing.Process]
        Processes to run ss at nodes
    runners: List[PingRunner]
    """
    runners = []
    if dependency:
        for key, timings in ping_schedules.items():
            src_ns = key[0]
            dst_ns = key[1]
            dst_addr = key[2]
            ping_runner = PingRunner(
                src_ns, dst_addr, timings[0], timings[1] - timings[0], dst_ns
            )
            runners.append(ping_runner)
    else:
        logger.warning("ping not found.")
    return runners


def progress_bar(stop_time, precision=1):
    """
    Show a progress bar from from 0 `units` to `stop_time`

    The time unit is decided by `precision` in seconds. It is
    1s by default.

    Parameters
    -----------
    stop_time : int
        The time needed 100% completion
    precision : int
        Time unit for updating progress bar. 1 second be default
    """
    try:
        print()
        for _ in tqdm(range(0, stop_time, precision), desc="Experiment Progress"):
            sleep(precision)
        print()
    except KeyboardInterrupt:
        logger.debug(
            "ProgressBar process received KeyboardInterrupt. Stopping it gracefully."
        )

    logger.info("Cleaning up all the spawned child processes...")


def cleanup():
    """
    Clean up
    """
    # Remove results of the experiment
    SsResults.remove_all_results()
    NetperfResults.remove_all_results()
    TcResults.remove_all_results()
    PingResults.remove_all_results()

    kill_processes()


# Helper methods
# pylint: disable=too-many-arguments
def _get_start_stop_time_for_ss(
    src_ns, dst_ns, dst_addr, start_t, stop_t, ss_schedules
):
    """
    Find the start time and stop time to run ss command in node `src_ns`
    to a `dst_addr`

    Parameters
    ----------
    src_ns: str
        ss run from `src_ns`
    dst_ns: str
        destination network namespace for ss
    dst_addr: str
        Destination address
    start_t: int
        Start time of ss command
    stop_t: int
        Stop time of ss command
    ss_schedules: list
        List with ss command schedules

    Returns
    -------
    List: Updated ss_schedules
    """
    if (src_ns, dst_ns, dst_addr) not in ss_schedules:
        ss_schedules[(src_ns, dst_ns, dst_addr)] = (start_t, stop_t)
    else:
        (min_start, max_stop) = ss_schedules[(src_ns, dst_ns, dst_addr)]
        ss_schedules[(src_ns, dst_ns, dst_addr)] = (
            min(min_start, start_t),
            max(max_stop, stop_t),
        )

    return ss_schedules



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Class to handles IS-IS related functionalities"""

import hashlib
from functools import partial
from nest.engine.dynamic_routing import run_isisd
from nest.routing.route_daemons import RoutingDaemonBase


class Isis(RoutingDaemonBase):
    """
    Handles IS-IS related functionalities.
    """

    def __init__(self, router_ns_id, interfaces, conf_dir, **kwargs):
        super().__init__(router_ns_id, interfaces, "isisd", conf_dir, **kwargs)

    def create_basic_config(self):
        """
        Creates a file with basic configuration for IS-IS.
        Use base `add_to_config` directly for more complex configurations
        """

        area_id = "00.0000.0000.0000.0000.0000.0000"
        system_id_hash = str(
            int(
                hashlib.sha256(
                    self.interfaces[0]
                    .address.get_addr(with_subnet=False)
                    .encode("utf-8")
                ).hexdigest(),
                16,
            )
        )[:12]
        system_id = ".".join(map("".join, zip(*[iter(system_id_hash)] * 4)))

        self.add_to_config(f"router isis {self.router_ns_id}")
        self.add_to_config("is-type level-1")
        self.add_to_config(f"net {area_id}.{system_id}.00")
        for interface in self.interfaces:
            self.add_to_config(f"interface {interface.id}")
            if self.ipv6:
                self.add_to_config(f" ipv6 router isis {self.router_ns_id}")
            else:
                self.add_to_config(f" ip router isis {self.router_ns_id}")

        if self.log_file is not None:
            self.add_to_config(f"log file {self.log_file}")

    def run(self):
        """
        Runs the isisd command
        """
        super().run(
            engine_func=partial(
                run_isisd, self.router_ns_id, self.conf_file, self.pid_file
            )
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to interfaces in topology"""

import logging
from nest.input_validator import input_validator
from nest.input_validator.metric import Bandwidth, Delay, Percentage
from nest.topology.veth_end import VethEnd
from nest.topology.ifb import Ifb
from nest.topology import Address
from nest.topology.network import Network
from nest import engine
import nest.config as config

# pylint: disable=cyclic-import
from nest import topology

# Max length of interface when Topology Map is disabled
# i.e. 'assign_random_names' is set to False in config
MAX_CUSTOM_NAME_LEN = 15

logger = logging.getLogger(__name__)

# TODO: Improve this module such that the below pylint disables are no
# longer required

# pylint: disable=too-many-instance-attributes
# pylint: disable=protected-access
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-public-methods


class Interface:
    """
    Abstraction for an network interface.

    Attributes
    ----------
    name : str
        User given name for the interface
    id : str
        This value is used by `engine` to create emulated interface
        entity
    node : Node
        `Node` which contains this `Interface`
    address : str/Address
        IP address assigned to this interface
    """

    @input_validator
    def __init__(self, interface_name: str):
        """
        Constructor of Interface.

        *Note*: The creation of Interface object does not
        create any devices in the interface.
        [See `create_veth_pair` method]

        Parameters
        ----------
        interface_name : str
            Name of the interface
        """

        self._veth_end = VethEnd(interface_name, None)
        self._ifb = None

        # Track bandwidth
        self._bandwidth = None

    @property
    def name(self):
        """Getter for name"""
        return self._veth_end.name

    @property
    def id(self):
        """Getter for id"""
        return self._veth_end.id

    @property
    def pair(self):
        """
        Get other pair for this interface (assuming veth)
        """
        return self._pair

    @pair.setter
    def pair(self, interface):
        """
        Setter for the other end of the interface that it is connected to

        Parameters
        ----------
        interface : Interface
            The interface to which this interface is connected to
        """
        self._pair = interface

    @property
    def node_id(self):
        """
        Getter for the `Node` associated
        with the interface
        """
        return self._veth_end.node_id

    @node_id.setter
    def node_id(self, node_id):
        """
        Setter for the `Node` associated
        with the interface

        Parameters
        ----------
        node : Node
            The node where the interface is to be installed
        """
        self._veth_end.node_id = node_id

    @property
    def ifb_id(self):
        """
        Getter for the id of the ifb of
        the interface
        """
        return self._ifb.id

    @property
    def subnet(self):
        """Getter for the subnet to which the address belongs to"""
        return self._veth_end.address.get_subnet()

    @property
    def address(self):
        """
        Getter for the address associated
        with the interface
        """
        return self._veth_end.address

    @address.setter
    @input_validator
    def address(self, address: Address):
        """
        Assigns IP address to an interface

        Parameters
        ----------
        address : Address or str
            IP address to be assigned to the interface
        """
        self._veth_end.address = address

    def enable_mpls(self):
        """
        Enables mpls input through the interface.
        Requires mpls kernel modules to be loaded.

        Run ``sudo modprobe mpls_iptunnel`` to load mpls modules.
        """
        self._veth_end.enable_mpls()

    def is_mpls_enabled(self):
        """
        Check if the interface has mpls enabled
        """
        return self._veth_end.is_mpls_enabled()

    @property
    def mtu(self):
        """
        Get the maximum transmit unit value for the interface
        """
        return self._veth_end.mtu

    @mtu.setter
    def mtu(self, mtu_value):
        """
        Set the maximum transmit unit value for the interface
        Default is 1500 bytes.
        """
        self._veth_end.mtu = mtu_value

    def get_address(self):
        """
        Getter for the address associated
        with the interface

        *NOTE*: Maintained since mentioned in NeST paper.
        """
        return self._veth_end.address

    @input_validator
    def set_address(self, address: Address):
        """
        Assigns IP address to an interface

        Parameters
        ----------
        address : Address or str
            IP address to be assigned to the interface

        *NOTE*: Maintained since mentioned in NeST paper.
        """
        self._veth_end.address = address

    def disable_ip_dad(self):
        """
        Disables Duplicate addresses Detection (DAD) for an interface.
        """
        self._veth_end.disable_ip_dad()

    def set_mode(self, mode):
        """
        Changes the mode of the interface

        Parameters
        ----------
        mode : string
            interface mode to be set
        """
        self._veth_end.set_mode(mode)

    def get_qdisc(self):
        """
        Note that this is the qdisc set inside
        the IFB.
        """
        if self._ifb is not None:
            for qdisc in self._ifb.qdisc_list:
                if qdisc.parent == "1:1" and qdisc.handle == "11:":
                    return qdisc
        return None

    def _create_and_mirred_to_ifb(self):
        """
        Creates a IFB for the interface so that a Qdisc can be
        installed on it
        Mirrors packets to be sent out of the interface first to
        itself (IFB)
        Assumes the interface has already invoked _set_structure()

        Parameters
        ----------
        """

        ifb_name = "ifb-" + self.name
        self._ifb = Ifb(ifb_name, self.node_id, self.id)

    @input_validator
    def set_bandwidth(self, bandwidth: Bandwidth):
        """
        Sets a minimum bandwidth for the interface
        It is done by adding a HTB qdisc and a rate parameter to the class

        Parameters
        ----------
        bandwidth : Bandwidth
            The minimum rate that has to be set in kbit
        """

        self._veth_end.set_structure()
        self._bandwidth = bandwidth.string_value

        bandwidth_parameter = {"rate": bandwidth.string_value}

        self._veth_end.change_class("htb", "1:", "1:1", **bandwidth_parameter)

    @input_validator
    def set_delay(self, delay: Delay):
        """
        Adds a delay to the link between two namespaces
        It is done by adding a delay in the interface

        Parameters
        ----------
        delay : Delay
            The delay to be added
        """
        # TODO: Make adding delay possible without bandwidth being set

        self._veth_end.set_structure()

        delay_parameter = {"delay": delay.string_value}

        self._veth_end.change_qdisc("11:", "netem", **delay_parameter)

    @input_validator
    def set_packet_corruption(
        self, corrupt_rate: Percentage, correlation_rate: Percentage = None
    ):
        """
        allows the emulation of random noise introducing an error in a
        random position for a chosen percent of packets.
        It is also possible to add a correlation.

        Parameters
        ----------
        corrupt_rate : str
            rate of the packets to be corrupted
        correlation_rate : str
            correlation between the corrupted packets
        """
        parsed_correlation_rate = (
            correlation_rate.string_value if correlation_rate else ""
        )

        self._veth_end.set_structure()

        corrupt_parameter = {
            "corrupt": corrupt_rate.string_value,
            "": parsed_correlation_rate,
        }

        self._veth_end.change_qdisc("11:", "netem", **corrupt_parameter)

    @input_validator
    def set_packet_loss(
        self, loss_rate: Percentage, correlation_rate: Percentage = None
    ):
        """
        adds an independent loss probability to the packets outgoing from
        the chosen network interface.
        It is also possible to add a correlation

        Parameters
        ----------
        loss_rate : str
            rate of the packets to be lost
        correlation_rate : str
            correlation between the lost packets
        """
        parsed_correlation_rate = (
            correlation_rate.string_value if correlation_rate else ""
        )

        self._veth_end.set_structure()

        loss_parameter = {"loss": loss_rate.string_value, "": parsed_correlation_rate}

        self._veth_end.change_qdisc("11:", "netem", **loss_parameter)

    def set_qdisc(self, qdisc, **kwargs):
        """
        Adds the Queueing algorithm to the interface

        Parameters
        ----------
        qdisc : string
            The Queueing algorithm to be added
        bandwidth :
            Link bandwidth
        """

        self._veth_end.set_structure()

        if self._ifb is None:
            self._create_and_mirred_to_ifb()

        if self._bandwidth is not None:
            self._ifb.set_bandwidth(self._bandwidth)

        self._ifb.delete_qdisc("11:")
        self._ifb.add_qdisc(qdisc, "1:1", "11:", **kwargs)

    @input_validator
    def set_attributes(
        self, bandwidth: Bandwidth, delay: Delay, qdisc: str = None, **kwargs
    ):
        """
        Add attributes bandwidth, delay and qdisc to interface

        Parameters
        ----------
        bandwidth : str/Bandwidth
            Packet outgoing rate
        delay : str/Delay
            Delay before packet is sent out
        qdisc : string
            The Queueing algorithm to be added to interface
            (Default value = None)
        """

        self.set_bandwidth(bandwidth)
        self.set_delay(delay)

        if qdisc is not None:
            self.set_qdisc(qdisc, **kwargs)

    def enable_offload(self, offload_name):
        """
        API for enabling offloads
        Parameters
        ----------
        offload_name : str
            The type of offload names that need to enable
        """

        self._veth_end.enable_offload(offload_name)

    def disable_offload(self, offload_name):
        """
        API for disabling offloads
        Parameters
        ----------
        offload_name : str
            The type of offload names that need to disable
        """

        self._veth_end.disable_offload(offload_name)

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"


def create_veth_pair(interface1_name, interface2_name):
    """
    Handle creation of Veth pairs between
    `interface1_name` and `interface2_name`

    Parameters
    ----------
    interface1_name : str
            Name of one of the interfaces to be connected
    interface2_name : str
            Name of the other interface to be connected
    """

    interface1 = Interface(interface1_name)
    interface2 = Interface(interface2_name)

    interface1.pair = interface2
    interface2.pair = interface1

    # Create the veth
    engine.create_veth(interface1.id, interface2.id)

    return (interface1, interface2)


@input_validator
def connect(
    node1: topology.Node,
    node2: topology.Node,
    interface1_name: str = "",
    interface2_name: str = "",
    network: Network = None,
):
    """
    Connects two nodes `node1` and `node2`.
    Creates two paired Virtual Ethernet interfaces (veth) and returns
    them as a 2-element tuple.
    The first interface belongs to `node1`, and the second interface
    belongs to `node2`.

    Parameters
    ----------
    node1 : Node
        First veth interface added in this node
    node2 : Node
        Second veth interface added in this node
    interface1_name : str
        Name of first veth interface
    interface2_name : str
        Name of second veth interface
    network : Network
        Object of the Network to add interfaces

    Returns
    -------
    (interface1, interface2)
        2 tuple of created (paired) veth interfaces. `interface1` is in
        `n1` and `interface2` is in `n2`.
    """
    # Number of connections between `node1` and `node2`, set to `None`
    # initially since it hasn't been computed yet
    connections = None

    # Check interface names
    if interface1_name == "":
        connections = _number_of_connections(node1, node2)
        interface1_name = _autogenerate_interface_name(node1, node2, connections)

    if interface2_name == "":
        if connections is None:
            connections = _number_of_connections(node1, node2)
        # pylint: disable=arguments-out-of-order
        interface2_name = _autogenerate_interface_name(node2, node1, connections)

    # Create 2 interfaces
    (interface1, interface2) = create_veth_pair(interface1_name, interface2_name)

    node1._add_interface(interface1)
    node2._add_interface(interface2)

    interface1.set_mode("UP")
    interface2.set_mode("UP")

    # Disabling Duplicate Address Detection(DAD) at the interfaces
    if config.get_value("disable_dad") is True:
        interface1.disable_ip_dad()
        interface2.disable_ip_dad()

    # The network parameter takes precedence over "global" network level
    if network is None:
        network = Network.current_network
    # Add the interfaces to the network if mentioned
    if network is not None:
        network.add_interface(interface1)
        network.add_interface(interface2)

    return (interface1, interface2)


def _autogenerate_interface_name(node1, node2, connections):
    """
    Auto-generate interface names based on respective node names
    and number of connections
    """
    interface_name = node1.name + "-" + node2.name + "-" + str(connections)

    if config.get_value("assign_random_names") is False:
        if len(interface_name) > MAX_CUSTOM_NAME_LEN:
            raise ValueError(
                f"Auto-generated device name {interface_name} is too long. "
                f"The length of name should not exceed 15 characters."
            )

    return interface_name


def _number_of_connections(node1, node2):
    """
    Return number of connections between the two nodes
    """
    connections = 0

    if len(node1.interfaces) > len(node2.interfaces):
        node1, node2 = node2, node1

    for interface in node1.interfaces:
        if interface.pair.node_id == node2.id:
            connections = connections + 1

    return connections



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
The base exception for all exceptions created in NeST
"""


class NestBaseException(Exception):
    """
    Base exception for all exceptions created in NeST
    """



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test Address sub-module in topology"""

import unittest
from nest.topology.address import Address, Subnet

# pylint: disable=missing-docstring


class TestTopologyAddress(unittest.TestCase):
    def setUp(self):
        self.addr = "2001:1:1:1443::1/64"
        self.addr_without_subnet = "2001:1:1:1443::1"
        self.default_addr = "DEFAULT"
        self.subnet_addr = "2001:1:1:1443::/64"

    def test_get_addr(self):
        address = Address(self.addr)
        self.assertEqual(address.get_addr(), self.addr)
        self.assertEqual(address.get_addr(with_subnet=False), self.addr_without_subnet)

        with self.assertRaises(ValueError):
            Address("2001:1:1:1443::g/64")  # invalid character 'g'

        with self.assertRaises(ValueError):
            Address("2001:1asdfasf")  # invalid string

        with self.assertRaises(ValueError):
            Address("2001:1:1:1443:::/64")  # invalid :::

        with self.assertRaises(ValueError):
            Address(":/64")  # invalid :

        with self.assertRaises(ValueError):
            Address("2001:1:1:1443::g000")  # invalid mask g000

        with self.assertRaises(ValueError):
            Address("gggg:1:1:1443::")  # invalid mask gggg

    def test_address_subnet(self):
        subnet_address = Address(self.subnet_addr)
        self.assertTrue(subnet_address.is_subnet())

        address = Address(self.addr)
        self.assertEqual(address.get_subnet(), self.subnet_addr)

        with self.assertRaises(ValueError):
            Subnet("2001:1:1:1443::1/64")  # 2001:1:1:1443::1/64 is not a subnet

    def test_subnet(self):
        subnet = Subnet(self.subnet_addr)

        self.assertEqual(subnet.get_next_addr().get_addr(), "2001:1:1:1443::1/64")
        self.assertEqual(subnet.get_next_addr().get_addr(), "2001:1:1:1443::2/64")
        self.assertEqual(subnet.get_next_addr().get_addr(), "2001:1:1:1443::3/64")

        with self.assertRaises(ValueError):
            Subnet("2001:1:1:1443::1/64").get_next_addr()

        with self.assertRaises(ValueError):
            Subnet(
                "2001:1:1:1443::/128"
            ).get_next_addr()  # no next subnet as the mask is 128

    def test_default(self):
        def_addr = Address("DEFAULT")

        with self.assertRaises(ValueError):
            def_addr.is_ipv6()  # raises error since 'default' is not a valid IPv4 or IPv6 address

        self.assertEqual("default", def_addr.get_addr())
        self.assertEqual(False, def_addr.is_subnet())

        with self.assertRaises(Exception):
            def_addr.get_subnet()  # no subnet for default

        with self.assertRaises(Exception):
            Subnet("DEFAULT")  # DEFAULT cannot be a subnet


if __name__ == "__main__":
    unittest.main()
